# -*- coding: utf-8 -*-
"""

matplotlib based plotting functions

@author: Jussi (jnu@iki.fi)
"""

from __future__ import division

from builtins import zip
from builtins import range
from builtins import object
import matplotlib
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.lines import Line2D
import matplotlib.gridspec as gridspec
import os.path as op
import numpy as np
import logging

from . import models
from . import numutils
from . import normaldata
from .trial import Trial, nexus_trial, Gaitcycle
from .stats import AvgTrial
from .config import cfg


logger = logging.getLogger(__name__)


def save_pdf(filename, fig):
    """ Save figure fig into pdf filename """
    try:
        logger.debug('writing %s' % filename)
        with PdfPages(filename) as pdf:
            pdf.savefig(fig)
    except IOError:
        raise IOError('Error writing %s, '
                      'check that file is not already open.' % filename)


def time_dist_barchart(values, stddev=None, thickness=.5,
                       color=None, interactive=True, stddev_bars=True,
                       plotvars=None):
    """ Multi-variable and multi-condition barchart plot.
    values dict is keyed as values[condition][var][context],
    given by e.g. get_c3d_analysis()
    stddev can be None or a dict keyed as stddev[condition][var][context].
    If no stddev for a given condition, set stddev[condition] = None
    plotvars gives variables to plot (if not all) and their order.
    """

    if interactive:
        import matplotlib.pyplot as plt
        fig = plt.figure()
    else:
        fig = Figure()
        canvas = FigureCanvasAgg(fig)

    def _plot_label(ax, rects, texts):
        """Plot a label inside each rect"""
        # FIXME: just a rough heuristic for font size
        fontsize = 6 if len(rects) >= 3 else 8
        for rect, txt in zip(rects, texts):
            ax.text(rect.get_width() * .0, rect.get_y() + rect.get_height()/2.,
                    txt, ha='left', va='center', size=fontsize)

    def _plot_oneside(vars_, context, col):
        """ Do the bar plots for given context and column """
        for ind, var in enumerate(vars_):
            # FIXME: adding subplot here is unnnecessary and triggers mpl
            # depreciation
            ax = fig.add_subplot(gs[ind, col])
            ax.axis('off')
            # may have several bars (conditions) per variable
            vals_this = [values[cond][var][context] for cond in conds]
            if not np.count_nonzero(~np.isnan(vals_this)):
                continue
            stddevs_this = ([stddev[cond][var][context] if stddev[cond]
                             else None for cond in conds])
            units_this = len(conds) * [units[ind]]
            ypos = np.arange(len(vals_this) * thickness, 0, -thickness)
            xerr = stddevs_this if stddev_bars else None
            rects = ax.barh(ypos, vals_this, thickness, align='edge',
                            color=color, xerr=xerr)
            # FIXME: set axis scale according to var normal values
            ax.set_xlim([0, 1.5 * max(vals_this)])
            texts = list()
            for val, std, unit in zip(vals_this, stddevs_this, units_this):
                if val == 0:
                    texts += ['']
                elif std:
                    texts += [u'%.2f ± %.2f %s' % (val, std, unit)]
                else:
                    texts += [u'%.2f %s' % (val, unit)]
            _plot_label(ax, rects, texts)
        # return the last set of rects for legend
        return rects

    if color is None:
        color = ['tab:orange', 'tab:green', 'tab:red', 'tab:brown',
                 'tab:pink', 'tab:gray', 'tab:olive']

    conds = values.keys()
    vals_1 = values[conds[0]]
    varsets = [set(values[cond].keys()) for cond in conds]
    # vars common to all conditions
    vars_common = set.intersection(*varsets)

    if plotvars is not None:
        # pick specified vars that appear in all of the conditions
        plotvars_set = set(plotvars)
        vars_ok = set.intersection(plotvars_set, vars_common)
        if plotvars_set - vars_ok:
            logger.warning('some conditions are missing variables: %s'
                           % (plotvars_set - vars_ok))
        # to preserve original order
        vars_ = [var for var in plotvars if var in vars_ok]
    else:
        vars_ = vars_common

    units = [vals_1[var]['unit'] for var in vars_]

    # 3 columns: bar, labels, bar
    gs = gridspec.GridSpec(len(vars_), 3, width_ratios=[1, 1/3., 1])

    # variable names into the center column
    for ind, (var, unit) in enumerate(zip(vars_, units)):
        textax = fig.add_subplot(gs[ind, 1])
        textax.axis('off')
        label = '%s (%s)' % (var, unit)
        textax.text(0, .5, label, ha='center', va='center')

    rects = _plot_oneside(vars_, 'Left', 0)
    rects = _plot_oneside(vars_, 'Right', 2)

    if len(conds) > 1:
        fig.legend(rects, conds, fontsize=7,
                   bbox_to_anchor=(.5, 0), loc="lower right",
                   bbox_transform=fig.transFigure)

    fig.add_subplot(gs[0, 0]).set_title('Left')
    fig.add_subplot(gs[0, 2]).set_title('Right')

    return fig


class Plotter(object):

    def __init__(self, layout=None, normaldata_files=None, interactive=True):
        """ Plot gait data.

        layout: list of lists
            Variables to be plotted. Each list is a row of variables.
            If None, must be set before plotting (use plotter.layout = layout,
            where plotter is a Plotter instance)
        normaldata_files: list
            Normal data files to read (.xlsx or .gcd). If None, read the files
            specified from config.
        interactive: bool
            If True, start the pyplot event loop to show the figure
            in a GUI. If False, do not import pyplot (this is needed for
            e.g. embedding in Qt).
        """
        matplotlib.style.use(cfg.plot.mpl_style)

        self.gridspec = None
        if layout:
            self.layout = layout
        else:
            self._layout = None

        self._normaldata_files = dict()
        self._normaldata = dict()
        if normaldata_files is None:
            for fn in cfg.general.normaldata_files:
                self.add_normaldata(fn)

        self.trial = None
        self.fig = None
        self._footer = None
        self.legendnames = list()
        self.modelartists = list()
        self.emgartists = list()
        self.nrows = 0
        self.ncols = 0
        self.interactive = interactive
        # if in interactive mode, create the figure later - otherwise we
        # can create it now, as the size does not matter
        self.fig = None if interactive else Figure()

    def add_normaldata(self, filename):
        """ Read the given normal data file if not already read. Take the
        data from the file into use (i.e. it will be used as the default
        normaldata for the plots) """
        if filename in self._normaldata_files:
            ndata = self._normaldata_files[filename]
        else:
            logger.debug('reading new normal data from %s' % filename)
            ndata = normaldata.read_normaldata(filename)
            self._normaldata_files[filename] = ndata
        logger.debug('updating normal data from %s' % filename)
        self._normaldata.update(ndata)

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, layout, plotheightratios=None, plotwidthratios=None):
        """ Set plot layout.

        plotheightratios: list
            Height ratios for the plots. Length must equal number of rows in
            the layout. If None, will be automatically computed.
        plotwidthratios: list
            Width ratios for the plots. Length must equal number of columns in
            the layout. If None, will be equal.
        """
        if (not isinstance(layout, list) or not all([isinstance(item, list)
           for item in layout])):
            raise ValueError('Layout must be a list of lists')

        self._layout = layout
        self.allvars = [item for row in layout for item in row]
        self.n_proper_vars = len([v for v in self.allvars if v is not None])
        self.nrows = len(layout)
        if self.nrows == 0:
            raise ValueError('No data to plot')
        self.ncols = len(layout[0])

        # compute figure width and height - only used for interactive figures
        self.figh = min(self.nrows*cfg.plot.inch_per_row + 1,
                        cfg.plot.maxh)
        self.figw = min(self.ncols*cfg.plot.inch_per_col,
                        cfg.plot.maxw)

        if plotheightratios is None:
            plotheightratios = self._plot_height_ratios()
        elif len(plotheightratios) != len(self.nrows):
            raise ValueError('n of height ratios must match n of rows')

        self.gridspec = gridspec.GridSpec(self.nrows, self.ncols,
                                          height_ratios=plotheightratios,
                                          width_ratios=plotwidthratios)

    def _create_interactive_figure(self):
        """ Create pyplot controlled figure """
        import matplotlib.pylab as plt
        # auto size fig according to n of subplots w, limit size
        self.fig = plt.figure(figsize=(self.figw, self.figh))

    def open_nexus_trial(self):
        self.trial = nexus_trial()

    def open_trial(self, source):
        self.trial = Trial(source)

    def _var_type(self, var):
        """ Helper to return variable type """
        # FIXME: better checks for analog vars
        if var is None:
            return None
        elif models.model_from_var(var):
            return 'model'
        elif var in ('model_legend', 'emg_legend'):
            return var
        # check whether it's a configured EMG channel or exists in the data
        # source (both are ok)
        elif (self.trial and self.trial.emg.is_channel(var) or var in
              cfg.emg.channel_labels):
            return 'emg'
        else:
            raise ValueError('Unknown variable %s' % var)

    def _plot_height_ratios(self):
        """ Automatically adjust height ratios, if they are not specified """
        plotheightratios = []
        for row in self._layout:
            # this should take into account any analog variable
            if all([self._var_type(var) == 'emg' for var in row]):
                plotheightratios.append(cfg.plot.analog_plotheight)
            else:
                plotheightratios.append(1)
        return plotheightratios

    def set_footer(self, txt):
        """ Set footer text for figure """
        if self._footer:
            self._footer.remove()
        self._footer = self.fig.text(0, 0, txt, fontsize=8, color='black',
                                     ha='left', va='bottom')

    def tight_layout(self):
        """ Customized tight layout """
        if self.gridspec is None:
            return
        logger.debug('setting tight layout')
        self.gridspec.tight_layout(self.fig)
        # space for main title
        top = (self.figh - cfg.plot.titlespace) / self.figh
        # decrease vertical spacing
        hspace = .6
        # self.gridspec.update(hspace=hspace)
        # self.gridspec.update(top=top)
        self.gridspec.update(top=top, hspace=hspace)

    def plot_trial(self, trial=None,
                   model_cycles=None,
                   emg_cycles=None,
                   plotheightratios=None,
                   plotwidthratios=None,  # FIXME: not used?
                   model_tracecolor=None,
                   model_linestyle='-',
                   model_alpha=1.0,
                   split_model_vars=True,
                   auto_match_model_cycle=True,
                   model_stddev=None,
                   x_axis_is_time=False,
                   match_pig_kinetics=True,
                   auto_match_emg_cycle=True,
                   linestyles_context=False,
                   toeoff_markers=None,
                   annotate_emg=True,
                   emg_tracecolor=None,
                   emg_alpha=None,
                   plot_model_normaldata=True,
                   plot_emg_normaldata=True,
                   plot_emg_rms=False,
                   sharex=True,
                   show=True,
                   superpose=False,
                   maintitle=None,
                   maintitleprefix=None,
                   add_zeroline=False,
                   legend_maxlen=10):

        """ Create plot of variables. Parameters:

            trial : Trial
                Trial to plot. If None, plot self.trial.
        model_cycles : list | dict of int |  dict of list | 'all' | None |
                       'forceplate'
                Gait cycles to plot. See trial.get_cycle() for options.
        emg_cycles : list | dict of int | int | dict of list | 'all' | None
                Same as above, applied to EMG variables.
        plotheightratios : list
                Force height ratios of subplot rows, e.g. [1 2 2 2] would
                make first row half the height of others.
        plotheightratios : list
                Force width ratios of subplot columns.
        model_tracecolor : Matplotlib colorspec
                Line color for model variables. If None, will be
                taken from config.
        model_linestyle : Matplotlib line style
                Line style for model variables. If None, will be
                taken from config.
        model_alpha : float
                Alpha value for model variable traces (0.0 - 1.0)
        split_model_vars: bool
                If model var does not have a leading context 'L' or 'R', a side
                will be prepended according to the gait cycle context.
                E.g. 'HipMoment' -> 'LHipMoment'.
                This allows convenient L/R overlay of e.g. kinematics variables
                by specifying the variable without context.
        auto_match_model_cycle: bool
                If True, the model variable will be plotted only for cycles
                whose context matches the context of the variable. E.g.
                'LHipMomentX' will be plotted only for left side cycles.
                If False, the variable will be plotted for all cycles.
        auto_match_emg_cycle: bool
                If True, the EMG channel will be plotted only for cycles
                whose context matches the context of the channel name. E.g.
                'LRec' will be plotted only for left side cycles.
                If False, the channel will be plotted for all cycles.
        model_stddev : None or dict
                Specifies 'standard deviation' for model variables. Can be
                used to plot e.g. confidence intervals, or stddev if plotting
                averaged data.
        x_axis_is_time: bool
                For unnormalized variables, whether to plot x axis in seconds
                (default) or in frames.
        match_pig_kinetics: bool
                If True, Plug-in Gait kinetics variables will be plotted only
                for cycles that begin with a valid forceplate strike.
        linestyles_context:
                Automatically select line style for model variables according
                to context (defined in config)
        emg_tracecolor : Matplotlib color
                Select line color for EMG variables. If None, will be
                automatically selected (defined in config)
        emg_alpha : float
                Alpha value for EMG traces (0.0 - 1.0)
        plot_model_normaldata : bool
                Whether to plot normal data for model variables.
        plot_emg_normaldata : bool
                Whether to plot normal data. Uses either default normal data
                (in site_defs) or the data given when creating the plotter
                instance.
        plot_emg_rms : bool | string
                Whether to plot EMG RMS superposed on the EMG signal.
                If 'rms_only', plot only RMS.
        maintitle : str
                Main title for the plot.
        maintitleprefix : str
                If maintitle is not set, title will be set to
                maintitleprefix + trial name.
        sharex : bool
                Link the x axes together (will affect zooming)
        superpose : bool
                If superpose=False, create new figure. Otherwise superpose
                on existing figure.
        show : bool
                Whether to show the plot after plotting is finished. Use
                show=False if overlaying multiple trials and call show()
                after finished. If interactive=False, this has no effect.
        add_zeroline : bool
                Add line on y=0

        """
        if trial is None and self.trial is None:
            raise ValueError('No trial, specify one or call open_trial()')
        elif trial is None:
            trial = self.trial

        if self._layout is None:
            raise ValueError('Please set layout before plotting')

        # figure creation
        # TODO: simplify
        if self.interactive:
            if superpose:
                if self.fig is None:
                    logger.debug('no figure to superpose on, creating new one')
                    self._create_interactive_figure()
                else:
                    logger.debug('using existing figure for superpose')
            else:
                # interactive, new figure
                logger.debug('creating new figure')
                self._create_interactive_figure()
        else:  # non interactive - figure should exist already
            if superpose:
                pass  # superposing on existing figure
            else:
                # reusing the existing figure - no superpose
                self.fig.clear()

        # automatically set title if in interactive mode
        if maintitle is None and self.interactive:
            if maintitleprefix is None:
                maintitleprefix = ''
            maintitle = maintitleprefix + trial.trialname

        # auto adjust plot heights
        if plotheightratios is None:
            plotheightratios = self._plot_height_ratios()
        elif len(plotheightratios) != len(self.nrows):
            raise ValueError('n of height ratios must match n of rows')
        plotaxes = []

        def _empty_artist():
            return matplotlib.patches.Rectangle((0, 0), 1, 1, fc="w",
                                                fill=False, edgecolor='none',
                                                linewidth=0)

        def _axis_annotate(ax, text):
            """ Annotate at center of axis """
            ax.annotate(text, xy=(.5, .5), xycoords='axes fraction',
                        ha="center", va="center")

        def _no_ticks_or_labels(ax):
            ax.tick_params(axis='both', which='both', bottom=False,
                           top=False, labelbottom=False, right=False,
                           left=False, labelleft=False)

        def _shorten_name(name, max_len=10):
            """Shorten overlong names for legend etc."""
            return name if len(name) <= max_len else '..'+name[-max_len+2:]

        def _handle_cycle_arg(cycles):
            """Process the cycle argument"""
            if isinstance(cycles, Gaitcycle):
                return [cycles]
            elif cycles is None or cycles == [None]:
                return [None]
            elif (isinstance(cycles, list) and all([isinstance(c, Gaitcycle)
                  for c in cycles])):
                    return cycles
            else:
                return trial.get_cycles(cycles)

        # set default values for vars
        if toeoff_markers is None:
            toeoff_markers = cfg.plot.toeoff_markers

        if emg_tracecolor is None:
            emg_tracecolor = cfg.plot.emg_tracecolor

        if emg_alpha is None:
            emg_alpha = cfg.plot.emg_alpha

        if model_cycles is None:
            model_cycles = cfg.plot.default_model_cycles

        if emg_cycles is None:
            emg_cycles = cfg.plot.default_emg_cycles

        model_cycles = _handle_cycle_arg(model_cycles)
        emg_cycles = _handle_cycle_arg(emg_cycles)

        if not (model_cycles or emg_cycles):
            raise ValueError('No matching gait cycles found in data')

        """
        if self.fig is None or not superpose:
            # auto size fig according to n of subplots w, limit size
            self.figh = min(self.nrows*cfg.plot.inch_per_row + 1,
                            cfg.plot.maxh)
            self.figw = min(self.ncols*cfg.plot.inch_per_col,
                            cfg.plot.maxw)
            logger.debug('new figure: width %.2f, height %.2f'
                         % (self.figw, self.figh))
            self.fig = plt.figure(figsize=(self.figw, self.figh))
            self.gridspec = gridspec.GridSpec(self.nrows, self.ncols,
                                              height_ratios=plotheightratios)
        """

        is_avg_trial = isinstance(trial, AvgTrial)

        for i, var in enumerate(self.allvars):

            if len(self.fig.axes) == len(self.allvars):
                # looks like all axes were created already (superpose)
                ax = self.fig.axes[i]
            else:
                if sharex and len(plotaxes) > 0:
                    ax = self.fig.add_subplot(self.gridspec[i],
                                              sharex=plotaxes[-1])
                else:
                    ax = self.fig.add_subplot(self.gridspec[i])

            var_type = self._var_type(var)
            if var_type is None:
                ax.axis('off')
                continue

            if var_type == 'model':
                model = models.model_from_var(var)
                for cycle in model_cycles:
                    # logger.debug('cycle %d-%d' % (cycle.start, cycle.end))
                    is_unnormalized = cycle.start is None
                    trial.set_norm_cycle(cycle)

                    if (split_model_vars and cycle.context + var
                       in model.varnames):
                        varname = cycle.context + var
                    else:
                        varname = var

                    # check for kinetics variable
                    kin_ok = True
                    if match_pig_kinetics and model.is_kinetic_var(var):
                        kin_ok = cycle.on_forceplate

                    # whether to plot or not
                    x_, data = trial.get_model_data(varname)
                    # FIXME: varname[0] == cycle.context may not apply to
                    # all model vars
                    if (data is not None and kin_ok and (is_unnormalized or
                        varname[0] == cycle.context or not
                         auto_match_model_cycle)):

                        # logger.debug('plotting data for %s' % varname)
                        x = (x_ / trial.framerate if is_unnormalized and
                             x_axis_is_time else x_)
                        # FIXME: cycle may not have context?
                        tcolor = (model_tracecolor if model_tracecolor
                                  else cfg.plot.model_tracecolors
                                  [cycle.context])
                        lstyle = (cfg.plot.model_linestyles
                                  [cycle.context] if linestyles_context else
                                  model_linestyle)
                        lines_ = ax.plot(x, data, tcolor, linestyle=lstyle,
                                         linewidth=cfg.plot.model_linewidth,
                                         alpha=model_alpha)
                        # generate picker events for line artist
                        # FIXME: also for other vars
                        for line_ in lines_:
                            line_.set_picker(1)
                            line_._trialname = trial.trialname
                            line_._cycle = cycle
                        # add toeoff marker for this cycle
                        if (not is_unnormalized and not is_avg_trial and
                           toeoff_markers):
                            toeoff = cycle.toeoffn
                            ax.axvline(toeoff, color=tcolor, linewidth=.5)
                        # tighten x limits
                        ax.set_xlim(x[0], x[-1])
                    else:
                        # logger.debug('not plotting data for %s' % varname)
                        if data is None:
                            logger.debug('(no data)')

                    # each cycle gets its own stddev plot (if data was found)
                    if (model_stddev is not None and not is_unnormalized and
                       data is not None):
                        if varname in model_stddev:
                            sdata = model_stddev[varname]
                            stdx = np.linspace(0, 100, sdata.shape[0])
                            ax.fill_between(stdx, data-sdata,
                                            data+sdata,
                                            color=cfg.plot.
                                            model_stddev_colors[cycle.context],
                                            alpha=cfg.plot.
                                            model_stddev_alpha)
                            # tighten x limits
                            ax.set_xlim(stdx[0], stdx[-1])

                    # set labels, ticks, etc. after plotting last cycle
                    if cycle == model_cycles[-1]:

                        yunit = model.units[varname]
                        if yunit == 'deg':
                            yunit = u'\u00B0'  # degree sign
                        ydesc = [s[:3] for s in model.ydesc[varname]]  # shorten
                        ylabel_ = '%s  %s  %s' % (ydesc[0], yunit, ydesc[1])
                        ax.set(ylabel=ylabel_)
                        ax.xaxis.label.set_fontsize(cfg.
                                                    plot.label_fontsize)
                        ax.yaxis.label.set_fontsize(cfg.
                                                    plot.label_fontsize)
                        subplot_title = model.varlabels[varname]

                        # add n of averages for AvgTrial
                        if is_avg_trial:
                            subplot_title += (' (avg of %d cycles)' %
                                              trial.n_ok[varname])

                        prev_title = ax.get_title()
                        if prev_title and prev_title != subplot_title:
                            subplot_title = prev_title + ' / ' + subplot_title
                        ax.set_title(subplot_title)
                        ax.title.set_fontsize(cfg.plot.title_fontsize)


                        ax.locator_params(axis='y', nbins=6)  # less tick marks
                        ax.tick_params(axis='both', which='major',
                                       labelsize=cfg.plot.ticks_fontsize)

                        if is_unnormalized and var in self.layout[-1]:
                            xlabel = 'Time (s)' if x_axis_is_time else 'Frame'
                            ax.set(xlabel=xlabel)
                            ax.xaxis.label.set_fontsize(cfg.
                                                        plot.label_fontsize)

                        if plot_model_normaldata and not is_unnormalized:
                            # normaldata vars are without preceding side
                            # this is a bit hackish
                            if varname[0].upper() in ['L', 'R']:
                                nvarname = varname[1:]
                            if nvarname in self._normaldata:
                                key = nvarname
                            else:
                                key = None
                            ndata = (self._normaldata[key] if key in
                                     self._normaldata else None)
                            if ndata is not None:
                                logger.debug('plotting model normaldata for %s'
                                             % varname)
                                normalx = np.linspace(0, 100, ndata.shape[0])
                                ax.fill_between(normalx, ndata[:, 0],
                                                ndata[:, 1],
                                                color=cfg.plot.
                                                model_normals_color,
                                                alpha=cfg.plot.
                                                model_normals_alpha)
                                # tighten x limits
                                ax.set_xlim(normalx[0], normalx[-1])

                        if add_zeroline:
                            ax.axhline(0, color='black', linewidth=.5)

            elif var_type == 'emg':
                # set title first, since we may end up not plotting the emg at
                # all (i.e for missing / disconnected channels)
                subplot_title = (cfg.emg.channel_labels[var] if
                                 var in cfg.emg.channel_labels
                                 else var)
                prev_title = ax.get_title()
                if prev_title and prev_title != subplot_title:
                    subplot_title = prev_title + ' / ' + subplot_title
                ax.set_title(subplot_title)
                ax.title.set_fontsize(cfg.plot.title_fontsize)

                for cycle in emg_cycles:
                    trial.set_norm_cycle(cycle)
                    is_unnormalized = cycle.start is None
                    try:
                        x_, data = trial.get_emg_data(var)
                    except KeyError:  # channel not found
                        _no_ticks_or_labels(ax)
                        if annotate_emg:
                            _axis_annotate(ax, 'not found')
                        break  # skip all cycles
                    if not trial.emg.status_ok(var):
                        _no_ticks_or_labels(ax)
                        if annotate_emg:
                            _axis_annotate(ax, 'disconnected')
                        break  # data no good - skip all cycles
                    x = (x_ / trial.analograte if is_unnormalized and
                         x_axis_is_time else x_ / 1.)

                    if is_unnormalized and not x_axis_is_time:
                        # analog -> frames
                        x /= trial.samplesperframe

                    if (is_unnormalized or var[0] == cycle.context or not
                       auto_match_emg_cycle):
                        # plot data and/or rms
                        if plot_emg_rms != 'rms_only':
                            ax.plot(x, data*cfg.plot.emg_multiplier,
                                    emg_tracecolor,
                                    linewidth=cfg.plot.emg_linewidth,
                                    alpha=emg_alpha)

                        if plot_emg_rms:
                            rms = numutils.rms(data, cfg.emg.rms_win)
                            ax.plot(x, rms*cfg.plot.emg_multiplier,
                                    emg_tracecolor,
                                    linewidth=cfg.plot.emg_rms_linewidth,
                                    alpha=emg_alpha)

                    if cycle == emg_cycles[-1]:
                        # last cycle; plot scales, titles etc.
                        ax.set(ylabel=cfg.plot.emg_ylabel)
                        ax.yaxis.label.set_fontsize(cfg.
                                                    plot.label_fontsize)
                        ax.locator_params(axis='y', nbins=4)
                        # tick font size
                        ax.tick_params(axis='both', which='major',
                                       labelsize=cfg.plot.ticks_fontsize)
                        ax.set_xlim(min(x), max(x))
                        ysc = [-cfg.plot.emg_yscale, cfg.plot.emg_yscale]
                        ax.set_ylim(ysc[0]*cfg.plot.emg_multiplier,
                                    ysc[1]*cfg.plot.emg_multiplier)

                        if (plot_emg_normaldata and not is_unnormalized and
                           var in cfg.emg.channel_normaldata):
                            # plot EMG normal bars
                            emgbar_ind = cfg.emg.channel_normaldata[var]
                            for k in range(len(emgbar_ind)):
                                inds = emgbar_ind[k]
                                ax.axvspan(inds[0], inds[1], alpha=cfg.
                                           plot.emg_normals_alpha,
                                           color=cfg.plot.
                                           emg_normals_color)

                        if is_unnormalized and var in self.layout[-1]:
                            xlabel = 'Time (s)' if x_axis_is_time else 'Frame'
                            ax.set(xlabel=xlabel)
                            ax.xaxis.label.set_fontsize(cfg.
                                                        plot.label_fontsize)

            elif var_type in ('model_legend', 'emg_legend'):
                # add current trial name to legend and update legend
                ax.set_axis_off()
                tr_name = _shorten_name(trial.trialname, legend_maxlen)
                leg_entry = (tr_name, trial.eclipse_data['DESCRIPTION'],
                             trial.eclipse_data['NOTES'])
                self.legendnames.append('%s   %s   %s' % leg_entry)

                if var_type == 'model_legend':
                    if linestyles_context:
                        # indicate line styles in legend if they are used
                        leg_ctxt_titles = ['line style for right foot',
                                           'line style for left foot']
                        leg_ctxt_artists = [Line2D((0, 1), (0, 0),
                                                   color='black',
                                                   linewidth=1,
                                                   linestyle=cfg.plot.
                                                   model_linestyles[ctxt])
                                            for ctxt in ['R', 'L']]
                    else:
                        leg_ctxt_titles = []
                        leg_ctxt_artists = []
                    # FIXME: model tracecolor arg is mandatory here
                    self.modelartists.append(Line2D((0, 1), (0, 0),
                                                    color=model_tracecolor,
                                                    linewidth=2))
                    ncol = (1 if len(self.legendnames) + len(leg_ctxt_titles)
                            < 5 else 2)
                    self.fig.legends = []
                    self.fig.legend(self.modelartists + leg_ctxt_artists,
                                    self.legendnames + leg_ctxt_titles,
                                    loc='lower left', ncol=ncol,
                                    prop={'size': cfg.plot.legend_fontsize},
                                    bbox_to_anchor=[.02, .02])

                else:  # emg_legend
                    self.emgartists.append(Line2D((0, 1), (0, 0),
                                           linewidth=2,
                                           color=emg_tracecolor))
                    ncol = 1 if len(self.legendnames) < 5 else 2
                    self.fig.legends = []
                    self.fig.legend(self.emgartists, self.legendnames,
                                    loc='lower left', ncol=ncol,
                                    prop={'size': cfg.plot.legend_fontsize},
                                    bbox_to_anchor=[.02, .02])

            plotaxes.append(ax)

        self.set_title(maintitle)
        #self.fig.tight_layout()
        #self.fig.subplots_adjust(left=0, bottom=0, right=1, top=1)

        if show and self.interactive:
            self.show()

        return self.fig

    def set_title(self, title):
        self.fig.suptitle(title, fontsize=cfg.plot.maintitle_fontsize,
                          fontweight="bold")
        self.tight_layout()  # update plot spacing

    def title_with_eclipse_info(self, trial=None, prefix=''):
        """ Create title: prefix + trial name + Eclipse description and
        notes """
        if trial is None:
            trial = self.trial
        desc = self.trial.eclipse_data['DESCRIPTION']
        notes = self.trial.eclipse_data['NOTES']
        maintitle = ('%s %s' % (prefix, self.trial.trialname) if prefix else
                     self.trial.trialname)
        maintitle += ' (%s)' % desc if desc else ''
        maintitle += ' (%s)' % notes if notes else ''
        return maintitle

    def show(self):
        # start the pyplot event loop to show the figure
        if self.interactive and self.fig is not None:
            logger.debug('starting pyplot event loop')
            import matplotlib.pyplot as plt
            plt.show()

    def create_pdf(self, pdf_name=None, trial=None, pdf_prefix='Nexus_plot',
                   sessionpath=None):
        """ Make a pdf out of the created figure.

        pdf_name: string
            Name of pdf file to create, without path. If not specified, Nexus
            trial name will be used.
        pdf_prefix: string
            Optional prefix for the name
        sessionpath: string
            Where to write the pdf. If not specified, written into the
            session directory of currently loaded trial.
        """
        if trial is None:
            trial = self.trial
        if not self.fig:
            raise ValueError('No figure to save!')
        if sessionpath is None:
            sessionpath = self.trial.sessionpath
        if not sessionpath:
            raise ValueError('Cannot get session path')
        # resize to A4
        # self.fig.set_size_inches([8.27, 11.69])
        if pdf_name:
            pdf_name = op.join(sessionpath, pdf_name)
        else:
            pdf_name = pdf_prefix + self.trial.trialname + '.pdf'
            pdf_name = op.join(sessionpath, pdf_name)
        if op.isfile(pdf_name):
            pass  # can prevent overwriting here
        try:
            logger.debug('writing %s' % pdf_name)
            with PdfPages(pdf_name) as pdf:
                pdf.savefig(self.fig)
        except IOError:
            raise IOError('Error writing %s, '
                          'check that file is not already open.' % pdf_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 03 14:54:34 2015

Copy trial videos to desktop under nexus_videos

@author: Jussi (jnu@iki.fi)
"""

import os
import os.path as op
import shutil
import itertools
import logging

from gaitutils import nexus, sessionutils, cfg, GaitDataError


logger = logging.getLogger(__name__)


def do_copy():

    nexus.check_nexus()

    dest_dir = op.join(op.expanduser('~'), 'Desktop', 'nexus_videos')
    if not op.isdir(dest_dir):
        os.mkdir(dest_dir)

    sessionpath = nexus.get_sessionpath()
    c3dfiles = sessionutils.get_c3ds(sessionpath, tags=cfg.eclipse.repr_tags,
                                     trial_type='dynamic')
    vidfiles = itertools.chain.from_iterable(sessionutils.get_trial_videos(c3d)
                                             for c3d in c3dfiles)
    vidfiles = list(vidfiles)
    if not vidfiles:
        raise GaitDataError('No video files found for representative trials')

    for vidfile in vidfiles:
        logger.debug('copying %s -> %s' % (vidfile, dest_dir))
        shutil.copy2(vidfile, dest_dir)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    do_copy()



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 17 14:41:31 2015

Stuff related to Python environment

@author: Jussi (jnu@iki.fi)
"""

import sys
import traceback
import subprocess
import psutil
import os.path as op
from pkg_resources import resource_filename

from .guiutils import error_exit


class GaitDataError(Exception):
    pass


def _count_script_instances(scriptname):
    """Count running instances of Python script"""
    nprocs = 0
    for proc in psutil.process_iter():
        try:
            cmdline = proc.cmdline()
            if len(cmdline) > 1:
                if 'python' in cmdline[0] and scriptname in cmdline[1]:
                    nprocs += 1
        # catch NoSuchProcess for procs that disappear inside loop
        except (psutil.AccessDenied, psutil.NoSuchProcess):
            continue
    return nprocs


def _git_autoupdate():
    """Hacky way to update repo to latest master, if running under git.
    Return True if update was ran, else False"""
    mod_dir = resource_filename('gaitutils', '')
    repo_dir = op.abspath(op.join(mod_dir, op.pardir))
    if op.isdir(op.join(repo_dir, '.git')):
        print('running git autoupdate')
        o = subprocess.check_output(['git', 'pull'], cwd=repo_dir)
        updated = 'pdating' in o  # XXX: fragile as hell
        if updated:
            print('autoupdate status: %s' % o)
            return True
        else:
            print('package already up to date')
            return False
    else:  # not a git repo
        print('%s is a git repo, not running autoupdate' % repo_dir)
        return False


def register_gui_exception_handler(full_traceback=False):
    """ Registers an exception handler that reports uncaught exceptions
    via GUI"""
    from .config import cfg

    def _my_excepthook(type_, value, tback):
        """ Custom exception handler for fatal (unhandled) exceptions:
        report to user via GUI and terminate. """
        # exception and message, but no traceback
        tbackstr = tback if full_traceback else ''
        msg = ''.join(traceback.format_exception(type_, value, tbackstr))
        error_exit(msg)
        # just the message (e.g. ValueError: "blah" -> "blah")
        # may sometimes be confusing, since type of exception is not printed
        # error_exit(value)
        #
        sys.__excepthook__(type_, value, tback)
        sys.exit()

    if cfg.general.gui_exceptions:
        sys.excepthook = _my_excepthook


def run_from_ipython():
    try:
        __IPYTHON__
        return True
    except NameError:
        return False



# -*- coding: utf-8 -*-
"""
Report related functions

@author: Jussi (jnu@iki.fi)
"""


from __future__ import division
from builtins import str
from past.builtins import basestring
import numpy as np
import dash
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State
import flask
from flask import request
from collections import OrderedDict
import logging
import os.path as op
import os
import subprocess
import ctypes
import base64
import io

import gaitutils
from gaitutils import (cfg, normaldata, models, layouts, GaitDataError,
                       sessionutils, numutils, videos)
from gaitutils.plot_plotly import plot_trials
from gaitutils.scripts import nexus_time_distance_vars

logger = logging.getLogger(__name__)


def convert_videos(vidfiles, check_only=False):
    """Convert video files using command and options defined in cfg.
    If check_only, return whether files were already converted.
    Instantly starts as many converter processes as there are files and
    returns. This has the disadvantage of potentially starting dozens of
    processes, causing slowdown.
    """
    CONV_EXT = '.ogv'  # extension for converted files
    if not isinstance(vidfiles, list):
        vidfiles = [vidfiles]
    convfiles = {vidfile: op.splitext(vidfile)[0] + CONV_EXT for vidfile
                 in vidfiles}
    converted = [op.isfile(fn) for fn in convfiles.values()]  # already done
    if check_only:
        return all(converted)

    # XXX: this disables Windows protection fault dialogs
    # needed since ffmpeg2theora may crash after conversion is complete (?)
    SEM_NOGPFAULTERRORBOX = 0x0002  # From MSDN
    ctypes.windll.kernel32.SetErrorMode(SEM_NOGPFAULTERRORBOX)

    vidconv_bin = cfg.general.videoconv_path
    vidconv_opts = cfg.general.videoconv_opts
    if not (op.isfile(vidconv_bin) and os.access(vidconv_bin, os.X_OK)):
        raise ValueError('Invalid video converter executable: %s'
                         % vidconv_bin)
    procs = []
    for vidfile, convfile in convfiles.items():
        if not op.isfile(convfile):
            # supply NO_WINDOW flag to prevent opening of consoles
            cmd = [vidconv_bin]+vidconv_opts.split()+[vidfile]
            cmd = [s.encode('iso-8859-1') for s in cmd]
            p = subprocess.Popen(cmd,
                                 stdout=None, creationflags=0x08000000)
            procs.append(p)
    return procs


def _make_dropdown_lists(options):
    """This takes a list of label/value dicts (with arbitrary type values)
    and returns list and dict. Needed since dcc.Dropdown can only take str
    values. identity is fed to dcc.Dropdown() and mapper is used for getting
    the actual values at the callback."""
    identity = list()
    mapper = dict()
    for option in options:
        di = {'label': option['label'], 'value': option['label']}
        if 'disabled' in option and option['disabled']:
            di['disabled'] = True
        identity.append(di)
        mapper[option['label']] = option['value']
    return identity, mapper


def _time_dist_plot(c3ds, sessions):
    cond_labels = [op.split(session)[-1] for session in sessions]
    fig = nexus_time_distance_vars._plot_trials(c3ds, cond_labels,
                                                interactive=False)
    buf = io.BytesIO()
    fig.savefig(buf, format='svg', bbox_inches='tight')
    buf.seek(0)
    return buf


# helper to shutdown flask server, see http://flask.pocoo.org/snippets/67/
def _shutdown_server():
    func = request.environ.get('werkzeug.server.shutdown')
    if func is None:
        raise RuntimeError('Not running with the Werkzeug Server')
    func()


def dash_report(info=None, sessions=None, tags=None, signals=None):
    """Returns dash app for web report.
    info: patient info
    sessions: list of session dirs
    tags: tags for dynamic gait trials
    signals: instance of ProgressSignals, used to send progress updates across
    threads
    """

    signals.progress.emit('Collecting trials...', 0)
    # relative width of left panel (1-12)
    # 3-session comparison uses narrower video panel
    # LEFT_WIDTH = 8 if len(sessions) == 3 else 7
    LEFT_WIDTH = 8
    VIDS_TOTAL_HEIGHT = 88  # % of browser window height
    # reduce to set, since there may be several labels for given id
    camera_labels = set(cfg.general.camera_labels.values())
    model_cycles = cfg.plot.default_model_cycles
    emg_cycles = cfg.plot.default_emg_cycles

    if len(sessions) < 1 or len(sessions) > 3:
        raise ValueError('Need a list of one to three sessions')

    is_comparison = len(sessions) > 1

    sessions_str = ' / '.join([op.split(s)[-1] for s in sessions])
    report_type = ('Single session report:' if len(sessions) == 1
                   else 'Comparison report:')
    report_name = '%s %s' % (report_type, sessions_str)

    info = info or sessionutils.default_info()

    # tags for dynamic trials
    # if doing a comparison, pick representative trials only
    dyn_tags = tags or (cfg.eclipse.repr_tags if is_comparison else
                        cfg.eclipse.tags)
    # will be shown in the menu for static trials
    static_tag = 'Static'

    age = None
    if info['hetu'] is not None:
        # compute subject age at session time
        session_dates = [sessionutils.get_session_date(session) for
                         session in sessions]
        ages = [numutils.age_from_hetu(info['hetu'], d) for d in
                session_dates]
        age = max(ages)

    # create Markdown text for patient info
    patient_info_text = '##### %s ' % (info['fullname'] if info['fullname']
                                       else 'Name unknown')
    if info['hetu']:
        patient_info_text += '(%s)' % info['hetu']
    patient_info_text += '\n\n'
    # if age:
    #     patient_info_text += 'Age at measurement time: %d\n\n' % age
    if info['report_notes']:
        patient_info_text += info['report_notes']

    # load normal data for gait models
    model_normaldata = dict()
    for fn in cfg.general.normaldata_files:
        ndata = normaldata.read_normaldata(fn)
        model_normaldata.update(ndata)
    if age is not None:
        age_ndata_file = normaldata.normaldata_age(age)
        if age_ndata_file:
            age_ndata = normaldata.read_normaldata(age_ndata_file)
            model_normaldata.update(age_ndata)

    # find the c3d files and build a nice dict
    c3ds = {session: dict() for session in sessions}
    for session in sessions:
        c3ds[session] = dict(dynamic=dict(), static=dict(), vid_only=dict())
        # collect dynamic trials for each tag
        for tag in dyn_tags:
            dyns = sessionutils.get_c3ds(session, tags=tag,
                                         trial_type='dynamic')
            if len(dyns) > 1:
                logger.warning('multiple tagged trials (%s) for %s' %
                               (tag, session))
            c3ds[session]['dynamic'][tag] = dyns[-1:]
        # require at least one dynamic for each session
        if not any(c3ds[session]['dynamic'][tag] for tag in dyn_tags):
            raise GaitDataError('No tagged dynamic trials found for %s' %
                                (session))
        # collect static trial (at most 1 per session)
        sts = sessionutils.get_c3ds(session, trial_type='static')
        if len(sts) > 1:
            logger.warning('multiple static trials for %s, using last one'
                           % session)
        c3ds[session]['static'][static_tag] = sts[-1:]
        # collect video-only dynamic trials
        for tag in cfg.eclipse.video_tags:
            dyn_vids = sessionutils.get_c3ds(session, tags=tag)
            if len(dyn_vids) > 1:
                logger.warning('multiple tagged video-only trials (%s) for %s'
                               % (tag, session))
            c3ds[session]['vid_only'][tag] = dyn_vids[-1:]

    # make Trial instances for all dynamic and static trials
    trials_dyn = list()
    trials_static = list()
    for session in sessions:
        for tag in dyn_tags:
            if c3ds[session]['dynamic'][tag]:
                tri = gaitutils.Trial(c3ds[session]['dynamic'][tag][0])
                trials_dyn.append(tri)
        if c3ds[session]['static'][static_tag]:
            tri = gaitutils.Trial(c3ds[session]['static']['Static'][0])
            trials_static.append(tri)

    # read some extra data from trials and create supplementary data
    tibial_torsion = dict()
    for tr in trials_dyn:
        # read tibial torsion for each trial and make supplementary traces
        # these will only be shown for KneeAnglesZ (knee rotation) variable
        tors = dict()
        tors['R'], tors['L'] = (tr.subj_params['RTibialTorsion'],
                                tr.subj_params['LTibialTorsion'])
        if tors['R'] is None or tors['L'] is None:
            logger.warning('could not read tibial torsion values from %s'
                           % tr.trialname)
            continue
        # include torsion info for all cycles; this is useful when plotting
        # isolated cycles
        cycs = tr.get_cycles(model_cycles)
        for cyc in cycs:
            tibial_torsion[cyc] = dict()
            for ctxt in tors:
                var_ = ctxt + 'KneeAnglesZ'
                tibial_torsion[cyc][var_] = dict()
                # x = % of gait cycle
                tibial_torsion[cyc][var_]['t'] = np.arange(101)
                # static tibial torsion value as function of x
                # convert radians -> degrees
                tibial_torsion[cyc][var_]['data'] = (np.ones(101) * tors[ctxt]
                                                     / np.pi * 180)
                tibial_torsion[cyc][var_]['label'] = ('Tib. tors. (%s) % s' %
                                                      (ctxt, tr.trialname))

    # find video files for all trials
    signals.progress.emit('Finding videos...', 0)
    # add camera labels for overlay videos
    # XXX: may cause trouble if labels already contain the string 'overlay'
    camera_labels_overlay = [lbl+' overlay' for lbl in camera_labels]
    camera_labels.update(camera_labels_overlay)

    # build dict of videos for given tag / camera label
    # videos will be listed in session order
    vid_urls = dict()
    all_tags = dyn_tags + [static_tag] + cfg.eclipse.video_tags
    for tag in all_tags:
        vid_urls[tag] = dict()
        for camera_label in camera_labels:
            vid_urls[tag][camera_label] = list()

    # collect all videos for given tag and camera, listed in session order
    for session in sessions:
        for trial_type in c3ds[session]:
            for tag, c3ds_this in c3ds[session][trial_type].items():
                if c3ds_this:
                    c3d = c3ds_this[0]  # only one c3d per tag and session
                    for camera_label in camera_labels:
                        overlay = 'overlay' in camera_label
                        real_camera_label = (camera_label[:camera_label.find(' overlay')]
                                             if overlay else camera_label)
                        vids_this = videos.get_trial_videos(c3d, camera_label=real_camera_label, vid_ext='.ogv', overlay=overlay)
                        if vids_this:
                            vid = vids_this[0]
                            url = '/static/%s' % op.split(vid)[1]
                            vid_urls[tag][camera_label].append(url)

    # build dcc.Dropdown options list for cameras and tags
    # list cameras which have videos for any tag
    opts_cameras = list()
    for camera_label in sorted(camera_labels):
        if any(vid_urls[tag][camera_label] for tag in all_tags):
            opts_cameras.append({'label': camera_label, 'value': camera_label})
    # list tags which have videos for any camera
    opts_tags = list()
    for tag in all_tags:
        if any(vid_urls[tag][camera_label] for camera_label in camera_labels):
            opts_tags.append({'label': '%s' % tag, 'value': tag})
    # add null entry in case we got no videos at all
    if not opts_tags:
        opts_tags.append({'label': 'No videos', 'value': 'no videos',
                          'disabled': True})

    # build dcc.Dropdown options list for the trials
    trials_dd = list()
    for tr in trials_dyn:
        trials_dd.append({'label': tr.name_with_description,
                          'value': tr.trialname})

    # in EMG layout, keep chs that are active in any of the trials
    signals.progress.emit('Reading EMG data', 0)
    try:
        emgs = [tr.emg for tr in trials_dyn]
        emg_layout = layouts.rm_dead_channels_multitrial(emgs,
                                                         cfg.layouts.std_emg)
    except GaitDataError:
        emg_layout = 'disabled'

    # FIXME: layouts into config?
    _layouts = OrderedDict([
            ('Patient info', 'patient_info'),
            ('Kinematics', cfg.layouts.lb_kinematics),
            ('Static kinematics', 'static_kinematics'),
            ('Static EMG', 'static_emg'),
            ('Kinematics + kinetics', cfg.layouts.lb_kin_web),
            ('Kinetics', cfg.layouts.lb_kinetics_web),
            ('EMG', emg_layout),
            ('Kinetics-EMG left', cfg.layouts.lb_kinetics_emg_l),
            ('Kinetics-EMG right', cfg.layouts.lb_kinetics_emg_r),
            ('Muscle length', cfg.layouts.musclelen),
            ('Time-distance variables', 'time_dist'),
            ])

    # pick desired single variables from model and append
    pig_singlevars_ = (models.pig_lowerbody.varlabels_noside.items() +
                       models.pig_lowerbody_kinetics.varlabels_noside.items())
    pig_singlevars = sorted(pig_singlevars_, key=lambda item: item[1])
    singlevars = OrderedDict([(varlabel, [[var]]) for var, varlabel in
                              pig_singlevars])
    _layouts.update(singlevars)

    # add supplementary data for normal layouts
    supplementary_default = dict()
    supplementary_default.update(tibial_torsion)

    dd_opts_multi_upper = list()
    dd_opts_multi_lower = list()

    for k, (label, layout) in enumerate(_layouts.items()):
        logger.debug('creating plot for %s' % label)
        signals.progress.emit('Creating plot: %s' % label, 100*k/len(_layouts))
        # for comparison report, include session info in plot legends and
        # use session specific line style
        trial_linestyles = 'session' if is_comparison else 'same'
        legend_type = ('short_name_with_tag' if is_comparison else
                       'tag_with_cycle')

        try:
            # special layout
            if isinstance(layout, basestring):
                if layout == 'time_dist':
                    # need c3ds in lists, one list for each session
                    c3ds_dyn = [[c3d for tag in dyn_tags for c3d in
                                 c3ds[session]['dynamic'][tag]]
                                for session in sessions]
                    buf = _time_dist_plot(c3ds_dyn, sessions)
                    encoded_image = base64.b64encode(buf.read())
                    graph_upper = html.Img(src='data:image/svg+xml;base64,{}'.
                                           format(encoded_image),
                                           id='gaitgraph%d' % k,
                                           style={'height': '100%'})
                    graph_lower = html.Img(src='data:image/svg+xml;base64,{}'.
                                           format(encoded_image),
                                           id='gaitgraph%d'
                                           % (len(_layouts)+k),
                                           style={'height': '100%'})

                elif layout == 'patient_info':
                    graph_upper = dcc.Markdown(patient_info_text)
                    graph_lower = graph_upper

                elif layout == 'static_kinematics':
                    layout_ = cfg.layouts.lb_kinematics
                    fig_ = plot_trials(trials_static, layout_,
                                       model_normaldata,
                                       model_cycles='unnormalized',
                                       emg_cycles=[],
                                       legend_type='short_name_with_cyclename',
                                       trial_linestyles=trial_linestyles)
                    graph_upper = dcc.Graph(figure=fig_, id='gaitgraph%d' % k,
                                            style={'height': '100%'})
                    graph_lower = dcc.Graph(figure=fig_, id='gaitgraph%d'
                                            % (len(_layouts)+k),
                                            style={'height': '100%'})

                elif layout == 'static_emg':
                    layout_ = cfg.layouts.std_emg
                    fig_ = plot_trials(trials_static, layout_,
                                       model_normaldata,
                                       model_cycles=[],
                                       emg_cycles='unnormalized',
                                       legend_type='short_name_with_cyclename',
                                       trial_linestyles=trial_linestyles)
                    graph_upper = dcc.Graph(figure=fig_, id='gaitgraph%d' % k,
                                            style={'height': '100%'})
                    graph_lower = dcc.Graph(figure=fig_, id='gaitgraph%d'
                                            % (len(_layouts)+k),
                                            style={'height': '100%'})

                # will be caught and menu item will be empty
                elif layout == 'disabled':
                    raise ValueError

                else:  # unrecognized layout; this is not caught by us
                    raise Exception('Unrecognized layout: %s' % layout)

            # regular gaitutils layout
            else:
                fig_ = plot_trials(trials_dyn, layout, model_normaldata,
                                   legend_type=legend_type,
                                   trial_linestyles=trial_linestyles,
                                   supplementary_data=supplementary_default)
                graph_upper = dcc.Graph(figure=fig_, id='gaitgraph%d' % k,
                                        style={'height': '100%'})
                graph_lower = dcc.Graph(figure=fig_, id='gaitgraph%d'
                                        % (len(_layouts)+k),
                                        style={'height': '100%'})

            dd_opts_multi_upper.append({'label': label, 'value': graph_upper})
            dd_opts_multi_lower.append({'label': label, 'value': graph_lower})

        except (ValueError, GaitDataError) as e:
            logger.warning('Failed to create plot for %s: %s' % (label, e))
            # insert the menu options but make them disabled
            dd_opts_multi_upper.append({'label': label, 'value': label,
                                        'disabled': True})
            dd_opts_multi_lower.append({'label': label, 'value': label,
                                        'disabled': True})
            continue

    opts_multi, mapper_multi_upper = _make_dropdown_lists(dd_opts_multi_upper)
    opts_multi, mapper_multi_lower = _make_dropdown_lists(dd_opts_multi_lower)

    def make_left_panel(split=True, upper_value='Kinematics',
                        lower_value='Kinematics'):
        """Make the left graph panels. If split, make two stacked panels"""

        # the upper graph & dropdown
        items = [
                    dcc.Dropdown(id='dd-vars-upper-multi', clearable=False,
                                 options=opts_multi,
                                 value=upper_value),

                    html.Div(id='div-upper', style={'height': '50%'}
                             if split else {'height': '100%'})
                ]

        if split:
            # add the lower one
            items.extend([
                            dcc.Dropdown(id='dd-vars-lower-multi',
                                         clearable=False,
                                         options=opts_multi,
                                         value=lower_value),

                            html.Div(id='div-lower', style={'height': '50%'})
                        ])

        return html.Div(items, style={'height': '80vh'})

    # create the app
    app = dash.Dash(__name__)
    # use local packaged versions of JavaScript libs etc. (no internet needed)
    app.css.config.serve_locally = True
    app.scripts.config.serve_locally = True

    # this is for generating the classnames in the CSS
    num2words = {1: 'one', 2: 'two', 3: 'three', 4: 'four', 5: 'five',
                 6: 'six', 7: 'seven', 8: 'eight', 9: 'nine', 10: 'ten',
                 11: 'eleven', 12: 'twelve'}
    classname_left = '%s columns' % num2words[LEFT_WIDTH]
    classname_right = '%s columns' % num2words[12-LEFT_WIDTH]

    app.layout = html.Div([  # row

            html.Div([  # left main div

                    html.H6(report_name),

                    dcc.Checklist(id='split-left',
                                  options=[{'label': 'Two panels',
                                            'value': 'split'}], values=[]),

                    # need split=True so that both panels are in initial layout
                    html.Div(make_left_panel(split=True), id='div-left-main')

                    ], className=classname_left),

            html.Div([  # right main div

                    dcc.Dropdown(id='dd-camera', clearable=False,
                                 options=opts_cameras,
                                 value='Front camera'),

                    dcc.Dropdown(id='dd-video-tag', clearable=False,
                                 options=opts_tags,
                                 value=opts_tags[0]['value']),

                    html.Div(id='videos'),

                    ], className=classname_right),

                     ], className='row')

    @app.callback(
            Output('div-left-main', 'children'),
            [Input('split-left', 'values')],
            [State('dd-vars-upper-multi', 'value')]
        )
    def update_panel_layout(split_panels, upper_value):
        split = 'split' in split_panels
        return make_left_panel(split, upper_value=upper_value)

    @app.callback(
            Output('div-upper', 'children'),
            [Input('dd-vars-upper-multi', 'value')]
        )
    def update_contents_upper_multi(sel_var):
        return mapper_multi_upper[sel_var]

    @app.callback(
            Output('div-lower', 'children'),
            [Input('dd-vars-lower-multi', 'value')]
        )
    def update_contents_lower_multi(sel_var):
        return mapper_multi_lower[sel_var]

    def _video_elem(title, url, max_height):
        """Create a video element with title"""
        if not url:
            return 'No video found'
        vid_el = html.Video(src=url, controls=True, loop=True, preload='auto',
                            title=title, style={'max-height': max_height,
                                                'max-width': '100%'})
        # return html.Div([title, vid_el])  # titles above videos
        return vid_el

    @app.callback(
            Output('videos', 'children'),
            [Input('dd-camera', 'value'),
             Input('dd-video-tag', 'value')]
        )
    def update_videos(camera_label, tag):
        """Create a list of video divs according to camera and tag selection"""
        if tag == 'no videos':
            return 'No videos found'
        vid_urls_ = vid_urls[tag][camera_label]
        if not vid_urls_:
            return 'No videos found'
        nvids = len(vid_urls_)
        max_height = str(int(VIDS_TOTAL_HEIGHT / nvids)) + 'vh'
        return [_video_elem('video', url, max_height) for url in vid_urls_]

    # add a static route to serve session data. be careful outside firewalls
    @app.server.route('/static/<resource>')
    def serve_file(resource):
        for session in sessions:
            filepath = op.join(session, resource)
            if op.isfile(filepath):
                return flask.send_from_directory(session, resource)
        return None

    # add shutdown method - see http://flask.pocoo.org/snippets/67/
    @app.server.route('/shutdown')
    def shutdown():
        logger.debug('Received shutdown request...')
        _shutdown_server()
        return 'Server shutting down...'

    return app



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 03 14:54:34 2015

Muscle len consistency plot from Nexus. Automatically picks trials based on
Eclipse description and defined tags.

@author: Jussi (jnu@iki.fi)
"""

import logging
import argparse
from itertools import cycle

from gaitutils import (Plotter, cfg, register_gui_exception_handler,
                       normaldata, GaitDataError, sessionutils, nexus)

logger = logging.getLogger(__name__)


def do_plot(sessionpath=None, tags=None, age=None, show=True, make_pdf=True):

    if sessionpath is None:
        sessionpath = nexus.get_sessionpath()

    tags = tags or cfg.eclipse.tags
    tagged_trials = sessionutils.get_c3ds(sessionpath, tags=tags,
                                          trial_type='dynamic')

    if not tagged_trials:
        raise GaitDataError('No marked trials found for current session')

    pl = Plotter()

    if age is not None:
        ndata = normaldata.normaldata_age(age)
        if ndata:
            pl.add_normaldata(ndata)

    pl.layout = cfg.layouts.overlay_musclelen

    linecolors = cfg.plot.overlay_colors
    ccolors = cycle(linecolors)

    for i, trialpath in enumerate(tagged_trials):
        logger.debug('plotting %s' % tagged_trials[i])
        pl.open_trial(tagged_trials[i])
        if i > len(linecolors):
            logger.warning('not enough colors for plot!')
        # only plot normaldata for last trial to speed up things
        plot_model_normaldata = (trialpath == tagged_trials[-1])
        pl.plot_trial(model_tracecolor=next(ccolors), linestyles_context=True,
                      toeoff_markers=False, add_zeroline=False, show=False,
                      maintitle='', superpose=True, sharex=False,
                      plot_model_normaldata=plot_model_normaldata)

    maintitle = ('Muscle length consistency plot, '
                 'session %s' % pl.trial.sessiondir)
    pl.set_title(maintitle)

    if show:
        pl.show()

    if make_pdf:
        pl.create_pdf('musclelen_consistency.pdf')

    return pl.fig


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('--tags', metavar='p', type=str, nargs='+',
                        help='strings that must appear in trial '
                        'description or notes')
    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG)
    register_gui_exception_handler()
    do_plot(tags=args.tags)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

Autoprocess single trial.
Note: does not use stats for event detection -> less accurate

@author: Jussi (jnu@iki.fi)
"""
from __future__ import absolute_import

import logging
import os.path as op

from gaitutils.scripts.nexus_autoprocess_session import (_do_autoproc,
                                                         _delete_c3ds)
from gaitutils import nexus, register_gui_exception_handler, GaitDataError


def autoproc_single():
    fn = nexus.get_trialname()
    if not fn:
        raise GaitDataError('No trial open in Nexus')
    fn += '.Trial.enf'
    enffiles = [op.join(nexus.get_sessionpath(), fn)]  # listify single enf
    _delete_c3ds(enffiles)
    _do_autoproc(enffiles)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    register_gui_exception_handler()
    autoproc_single()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

Script to create the comparison pdf report.
Note: specific to the Helsinki gait lab.


@author: Jussi (jnu@iki.fi)
"""
from __future__ import absolute_import

import logging
import os.path as op
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from collections import defaultdict

from gaitutils import (Plotter, cfg, register_gui_exception_handler, layouts,
                       numutils, normaldata, nexus, sessionutils)
from PyQt5 import QtWidgets
from gaitutils.scripts import (nexus_kin_consistency,
                               nexus_time_distance_vars)

logger = logging.getLogger(__name__)

sort_field = 'NOTES'  # sort trials by the given Eclipse key
repr_tags = cfg.eclipse.repr_tags
page_size = (11.69, 8.27)  # report page size


def _add_footer(fig, txt):
    fig.text(0, 0, txt, fontsize=8, color='black', ha='left', va='bottom')


def _add_header(fig, txt):
    fig.text(0, 1, txt, fontsize=8, color='black', ha='left', va='top')


def _savefig(pdf, fig, header=None, footer=None):
    """add header/footer into page and save as A4"""
    if fig is None:
        return
    if header is not None:
        _add_header(fig, header)
    if footer is not None:
        _add_footer(fig, footer)
    fig.set_size_inches(page_size[0], page_size[1])
    pdf.savefig(fig)


def do_plot(sessions, pdfpath=None, pages=None):
    """ Do a quick comparison report between sessions """

    if pages is None:
        # if no pages specified, do them all
        pages = defaultdict(lambda: True)
    else:
        if not any(pages.values()):
            raise Exception('No pages to print')

    sessions_str = u' vs. '.join([op.split(s)[-1] for s in sessions])

    # make header page
    fig_hdr = plt.figure()
    ax = plt.subplot(111)
    plt.axis('off')
    title_txt = 'HUS Liikelaboratorio\n'
    title_txt += u'Kävelyanalyysin vertailuraportti\n'
    title_txt += '\n'
    title_txt += sessions_str
    ax.text(.5, .8, title_txt, ha='center', va='center', weight='bold',
            fontsize=14)

    fig_timedist_cmp = (nexus_time_distance_vars.
                        do_comparison_plot(sessions, tags=repr_tags,
                                           show=False))

    fig_kin_cmp = nexus_kin_consistency.do_plot(sessions, tags=repr_tags,
                                                session_styles=True,
                                                show=False)

    if pdfpath is None:
        pdfpath = QtWidgets.QFileDialog.getSaveFileName(None,
                                                        'Save PDF',
                                                        sessions[0],
                                                        '*.pdf')[0]
    if pdfpath:

        header = u'Comparison %s' % sessions_str
        logger.debug('creating multipage comparison pdf %s' % pdfpath)
        with PdfPages(pdfpath) as pdf:
            _savefig(pdf, fig_hdr)
            _savefig(pdf, fig_timedist_cmp, header)
            _savefig(pdf, fig_kin_cmp, header)

    # close all created figures, otherwise they'll pop up on next show() call
    plt.close('all')


if __name__ == '__main__':
    raise Exception('Not runnable yet (need a way to specify sessions)')
    logging.basicConfig(level=logging.DEBUG)
    register_gui_exception_handler()
    do_plot()



# -*- coding: utf-8 -*-
"""
Created on Fri Mar  1 11:11:33 2019

@author: hus20664877
"""

from PyQt5 import QtCore, uic, QtWidgets
from pkg_resources import resource_filename
import os.path as op
import ast

from .. import nexus, GaitDataError, cfg


def qt_message_dialog(msg):
    """ Show message with an 'OK' button. """
    dlg = QtWidgets.QMessageBox()
    dlg.setWindowTitle('Message')
    dlg.setText(msg)
    dlg.addButton(QtWidgets.QPushButton('Ok'),
                  QtWidgets.QMessageBox.YesRole)
    dlg.exec_()


def qt_yesno_dialog(msg):
    """ Show message with 'Yes' and 'No buttons, return role accordingly """
    dlg = QtWidgets.QMessageBox()
    dlg.setWindowTitle('Confirm')
    dlg.setText(msg)
    dlg.addButton(QtWidgets.QPushButton('Yes'),
                  QtWidgets.QMessageBox.YesRole)
    dlg.addButton(QtWidgets.QPushButton('No'),
                  QtWidgets.QMessageBox.NoRole)
    dlg.exec_()
    return dlg.buttonRole(dlg.clickedButton())


def qt_dir_chooser():
    """Selector dialog to select dir (or multiple dirs)."""
    # native dialog - single dir only
    return [QtWidgets.QFileDialog.getExistingDirectory(None, 'Select session')]
    # non-native dialog - multiple dirs. a bit messy, currently not in use
    file_dialog = QtWidgets.QFileDialog()
    file_dialog.setFileMode(QtWidgets.QFileDialog.DirectoryOnly)
    file_dialog.setOption(QtWidgets.QFileDialog.DontUseNativeDialog,
                          True)
    file_view = file_dialog.findChild(QtWidgets.QListView, 'listView')
    if file_view:
        file_view.setSelectionMode(QtWidgets.QAbstractItemView.
                                   MultiSelection)
    f_tree_view = file_dialog.findChild(QtWidgets.QTreeView)
    if f_tree_view:
        f_tree_view.setSelectionMode(QtWidgets.QAbstractItemView.
                                     MultiSelection)
    return file_dialog.selectedFiles() if file_dialog.exec_() else []


class OptionsDialog(QtWidgets.QDialog):
    """ Display a tabbed dialog for changing gaitutils options """

    def __init__(self, parent, default_tab=0):
        super(self.__class__, self).__init__(parent)
        # load user interface made with designer
        uifile = resource_filename('gaitutils', 'gui/options_dialog.ui')
        uic.loadUi(uifile, self)
        #self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        # add some buttons to the standard button box
        loadButton = QtWidgets.QPushButton('Load...')
        self.buttonBox.addButton(loadButton,
                                 QtWidgets.QDialogButtonBox.ActionRole)
        loadButton.clicked.connect(self.load_config_dialog)
        saveButton = QtWidgets.QPushButton('Save...')
        self.buttonBox.addButton(saveButton,
                                 QtWidgets.QDialogButtonBox.ActionRole)
        saveButton.clicked.connect(self.save_config_dialog)

        # show page
        self.tabWidget.setCurrentIndex(default_tab)

        """ Collect config widgets into a dict of dict. First key is tab
        (same as config category, e.g. autoproc), second key is widget name """
        self.cfg_widgets = dict()
        for page in [self.tabWidget.widget(n) for n in
                     range(self.tabWidget.count())]:
            pname = page.objectName()
            self.cfg_widgets[pname] = dict()
            for w in page.findChildren(QtWidgets.QWidget):
                wname = w.objectName()
                if wname[:4] == 'cfg_':  # config widgets are specially named
                    self.cfg_widgets[pname][wname] = w
        self._update_widgets()
        self.homedir = op.expanduser('~')

    def load_config_dialog(self):
        """ Bring up load dialog and load selected file. """
        global cfg
        fout = QtWidgets.QFileDialog.getOpenFileName(self,
                                                     'Load config file',
                                                     self.homedir,
                                                     'Config files (*.cfg)')
        # TODO : filedialog set filter -> PyQt5.QtCore.QDir.Hidden?
        fname = fout[0]
        if fname:
            # TODO: check for errors on config read
            # cfg.load_default()  TODO: load defaults before loading cfg file?
            cfg.read(fname)
            self._update_widgets()

    def save_config_dialog(self):
        """ Bring up save dialog and save data. """
        global cfg
        res, txt = self._check_widget_inputs()
        if not res:
            qt_message_dialog('Invalid input: %s\nPlease fix before saving'
                              % txt)
        else:
            fout = QtWidgets.QFileDialog.getSaveFileName(self,
                                                         'Save config file',
                                                         self.homedir,
                                                         'Config files '
                                                         '(*.cfg)')
            fname = fout[0]
            if fname:
                self.update_cfg()
                cfg.write_file(fname)

    def _update_widgets(self):
        """ Update config widgets according to current cfg """
        for section in self.cfg_widgets:
            for wname, widget in self.cfg_widgets[section].items():
                item = wname[4:]
                cfgval = getattr(getattr(cfg, section), item)
                if str(cfgval) != str(self._getval(widget)):
                    self._setval(widget, cfgval)  # set using native type
                if isinstance(widget, QtWidgets.QLineEdit):
                    widget.setCursorPosition(0)  # show beginning of line

    def _check_widget_inputs(self):
        """ Check widget inputs. Currently only QLineEdits are checked for
        eval - ability """
        for section in self.cfg_widgets:
            for widget in self.cfg_widgets[section].values():
                if isinstance(widget, QtWidgets.QLineEdit):
                    txt = widget.text()
                    try:
                        ast.literal_eval(txt)
                    except (SyntaxError, ValueError):
                        return (False, txt)
        return (True, '')

    def _getval(self, widget):
        """ Universal value getter that takes any type of config widget.
        Returns native types, except QLineEdit input is auto-evaluated """
        if (isinstance(widget, QtWidgets.QSpinBox) or
           isinstance(widget, QtWidgets.QDoubleSpinBox)):
            return widget.value()
        elif isinstance(widget, QtWidgets.QCheckBox):
            return bool(widget.checkState())
        elif isinstance(widget, QtWidgets.QComboBox):
            # convert to str to avoid writing out unicode repr() into config
            # files unnecessarily
            return str(widget.currentText())
        elif isinstance(widget, QtWidgets.QLineEdit):
            # Directly eval lineEdit contents. This means that string vars
            # must be quoted in the lineEdit.
            txt = widget.text()
            return ast.literal_eval(txt) if txt else None
        else:
            raise Exception('Unhandled type of config widget')

    def _setval(self, widget, val):
        """ Universal value setter that takes any type of config widget.
        val must match widget type, except for QLineEdit that can take
        any type, which will be converted to its repr """
        if (isinstance(widget, QtWidgets.QSpinBox) or
           isinstance(widget, QtWidgets.QDoubleSpinBox)):
            widget.setValue(val)
        elif isinstance(widget, QtWidgets.QCheckBox):
            widget.setCheckState(2 if val else 0)
        elif isinstance(widget, QtWidgets.QComboBox):
            idx = widget.findText(val)
            if idx >= 0:
                widget.setCurrentIndex(idx)
            else:
                raise ValueError('Tried to set combobox to invalid value.')
        elif isinstance(widget, QtWidgets.QLineEdit):
            widget.setText(repr(val))
        else:
            raise Exception('Unhandled type of config widget')

    def update_cfg(self):
        """ Update cfg according to current dialog settings """
        global cfg
        for section in self.cfg_widgets.keys():
            for wname, widget in self.cfg_widgets[section].items():
                item = wname[4:]
                widgetval = self._getval(widget)
                cfgval = getattr(getattr(cfg, section), item)
                if widgetval != cfgval:
                    cfg[section][item] = repr(widgetval)

    def accept(self):
        """ Update config and close dialog, if widget inputs are ok. Otherwise
        show an error dialog """
        res, txt = self._check_widget_inputs()
        if res:
            self.update_cfg()
            self.done(QtWidgets.QDialog.Accepted)  # or call superclass accept
        else:
            qt_message_dialog("Invalid input: %s" % txt)


class ChooseSessionsDialog(QtWidgets.QDialog):
    """Display a dialog for picking sessions"""

    def __init__(self, min_sessions=1, max_sessions=3):
        super(self.__class__, self).__init__()
        # load user interface made with designer
        uifile = resource_filename('gaitutils', 'gui/web_report_sessions.ui')
        uic.loadUi(uifile, self)
        #self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.btnBrowseSession.clicked.connect(self.add_session)
        self.btnAddNexusSession.clicked.connect(lambda: self.
                                                add_session(from_nexus=True))
        self.btnClearAll.clicked.connect(self.listSessions.clear)
        self.btnClearCurrent.clicked.connect(self.listSessions.rm_current_item)
        self.max_sessions = max_sessions
        self.min_sessions = min_sessions

    def add_session(self, from_nexus=False):
        if len(self.sessions) == self.max_sessions:
            qt_message_dialog('You can specify maximum of %d sessions' %
                              self.max_sessions)
            return
        if from_nexus:
            try:
                dirs = [nexus.get_sessionpath()]
            except GaitDataError:
                qt_message_dialog('Cannot get session path from Nexus')
                return
        else:
            dirs = qt_dir_chooser()
        dirs = [op.normpath(d) for d in dirs]
        if dirs:
            for dir_ in dirs:
                if dir_ in self.sessions:
                    qt_message_dialog('Session %s already loaded' % dir_)
                elif dir_:
                    self.listSessions.add_item(dir_, data=dir_)

    @property
    def sessions(self):
        return [item.userdata for item in self.listSessions.items]

    def accept(self):
        if len(self.sessions) < self.min_sessions:
            qt_message_dialog('Please select at least %d session%s' %
                              (self.min_sessions,
                               's' if self.min_sessions > 1 else ''))
        else:
            self.done(QtWidgets.QDialog.Accepted)




# -*- coding: utf-8 -*-
"""
Created on Fri Sep 23 10:59:28 2016

Eclipse (database) hacks.

@author: Jussi (jnu@iki.fi)
"""
from builtins import str
from builtins import object
import logging
import io
import configobj
from configobj import ConfigObj
from collections import defaultdict, OrderedDict

from .numutils import isint
from .envutils import GaitDataError

logger = logging.getLogger(__name__)


class FileFilter(object):
    """ File-like class that will replace strings according to the replace
    dict below. This is needed for prefiltering before configobj parsing,
    since configobj will not tolerate lines with neither key nor value
    (which seem to occasionally appear in Eclipse files).
    Also deduplicates successive identical lines for same reason """

    replace = {'\n=\n': '\n'}

    def __init__(self, fname):
        self.fname = fname

    def read(self):
        """ ConfigObj seems to use only this method """
        # Eclipse switched from latin-1 to utf-8 at some point...
        try:
            self.fp = io.open(self.fname, encoding='utf8')
            data = self.fp.read()
        except UnicodeDecodeError:
            logger.warning('Cannot interpret %s as utf-8, trying latin-1' %
                           self.fname)
            self.fp = io.open(self.fname, encoding='latin-1')
            data = self.fp.read()
        # filter
        for val, newval in FileFilter.replace.items():
            data = data.replace(val, newval)
        # rm subsequent duplicate lines - a bit cryptic
        data = '\n'.join(list(OrderedDict.fromkeys(data.split('\n'))))
        return data

    def close(self):
        self.fp.close()


def _enf_reader(fname_enf):
    """ Return enf reader """
    fp = FileFilter(fname_enf)
    # do not listify comma-separated values
    # logger.debug('loading %s' % fname_enf)
    try:
        cp = ConfigObj(fp, encoding='utf8', list_values=False,
                       write_empty_values=True)
    except configobj.ParseError:
        raise GaitDataError('Cannot parse config file %s' % fname_enf)
    if 'TRIAL_INFO' not in cp.sections:
        raise GaitDataError('No trial info in .enf file')
    return cp


def get_eclipse_keys(fname_enf, return_empty=False):
    """ Read key/value pairs from ENF file into a dict. Only keys in the
    TRIAL_INFO section will be read. Return keys without value if
    return_empty.
    """
    di = defaultdict(lambda: u'')
    cp = _enf_reader(fname_enf)
    di.update({key: val for key, val in cp['TRIAL_INFO'].items()
               if val != '' or return_empty})
    return di


def eclipse_fp_keys(eclipse_keys):
    """ Filter that returns Eclipse forceplate keys/values as a dict """
    return {key: val for key, val in eclipse_keys.items()
            if key[:2] == 'FP' and len(key) == 3 and isint(key[2])}


def set_eclipse_keys(fname_enf, eclipse_dict, update_existing=False):
    """ Set key/value pairs in an ENF file. If update_existing is True,
    overwrite existing keys. Keys will be written into the TRIAL_INFO section.
    """
    cp = _enf_reader(fname_enf)
    did_set = False
    for key, val in eclipse_dict.items():
        if key not in cp['TRIAL_INFO'].keys() or update_existing:
            cp['TRIAL_INFO'][key] = val
            did_set = True
    if did_set:
        logger.debug('writing %s' % fname_enf)
        out = cp.write()  # output the config lines
        # result is utf8, but needs to be converted to unicode type for write
        outu = [str(line+'\n', encoding='utf8') for line in out]
        with io.open(fname_enf, 'w', encoding='utf8') as fp:
            fp.writelines(outu)
    else:
        logger.debug('did not set any keys')



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 17 14:41:31 2015

Class for reading EMG

@author: Jussi (jnu@iki.fi)
"""


from __future__ import division
from builtins import object
import numpy as np
from scipy import signal
import logging

from . import read_data
from .config import cfg

logger = logging.getLogger(__name__)


class EMG(object):
    """ Class for handling EMG data. """

    def __init__(self, source):
        logger.debug('new EMG instance from %s' % source)
        self.source = source
        # order of Butterworth filter
        self.buttord = 5
        self.passband = cfg.emg.passband
        self.linefreq = cfg.emg.linefreq
        self.data = None

    def __getitem__(self, item):
        """ Return data for a channel (filtered if self.passband is set).
        Uses name matching: if the specified channel is not found in the data,
        partial name matches are considered and data for the shortest match is
        returned. For example, 'LGas' could be mapped to 'Voltage.LGas8' """
        if item is None or len(item) < 2:
            raise KeyError('Invalid channel name')
        if self.data is None:
            self.read()
        matches = [x for x in self.data if x.find(item) >= 0]
        if len(matches) == 0:
            raise KeyError('No matching channel for %s' % item)
        else:
            ch = min(matches, key=len)  # choose shortest matching name
        if len(matches) > 1:
            logger.warning('multiple channel matches for %s: %s -> %s' %
                           (item, matches, ch))
        data = self.data[ch]
        return self.filt(data, self.passband) if self.passband else data

    def is_channel(self, item):
        """ Convenience to see whether a channel exists in the data """
        try:
            self[item]
            return True
        except KeyError:
            return False

    def status_ok(self, item):
        """ Returns True for existing channel with signal ok, False
        otherwise """
        return self.is_channel(item) and self._is_valid_emg(self[item])

    def _is_valid_emg(self, y):
        """ Check whether channel contains a valid EMG signal. Usually invalid
        signal can be identified by the presence of large powerline (harmonics)
        compared to broadband signal. Cause is typically disconnected/badly
        connected electrodes.
        TODO: should use multiple-zero IIR notch filter """
        # bandwidth of broadband signal. should be less than dist between
        # the powerline harmonics
        broadband_bw = 30
        power_bw = 4  # width of power line peak detector (bandpass)
        nharm = 3  # number of harmonics to detect
        # detect the 50 Hz harmonics
        linefreqs = (np.arange(nharm+1)+1) * self.linefreq
        intvar = 0
        for f in linefreqs:
            intvar += np.var(self.filt(y, [f-power_bw/2.,
                                           f+power_bw/2.])) / power_bw
        # broadband signal
        band = [self.linefreq+10, self.linefreq+10+broadband_bw]
        emgvar = np.var(self.filt(y, band)) / broadband_bw
        intrel = 10*np.log10(intvar/emgvar)
        return intrel < cfg.emg.max_interference

    def filt(self, y, passband):
        """ Filter given data y to passband, e.g. [1, 40].
        Passband is given in Hz. None for no filtering.
        Implemented as pure lowpass, if highpass freq = 0 """
        if passband is None:
            return y
        passbandn = 2 * np.array(passband) / self.sfrate
        if passbandn[0] > 0:  # bandpass
            b, a = signal.butter(self.buttord, passbandn, 'bandpass')
        else:  # lowpass
            b, a = signal.butter(self.buttord, passbandn[1])
        yfilt = signal.filtfilt(b, a, y)
        # yfilt = yfilt - signal.medfilt(yfilt, 21)
        return yfilt

    def read(self):
        meta = read_data.get_metadata(self.source)
        logger.debug('reading EMG from %s' % meta['trialname'])
        self.sfrate = meta['analograte']
        emgdi = read_data.get_emg_data(self.source)
        self.data = emgdi['data']
        self.t = emgdi['t']



# -*- coding: utf-8 -*-
"""

Non-Qt GUI stuff (Windows native dialogs etc.)

@author: Jussi (jnu@iki.fi)
"""

from builtins import str
import ctypes
import sys
import logging

logger = logging.getLogger(__name__)


def error_exit(message):
    """ Custom error handler """
    ctypes.windll.user32.MessageBoxA(0, message.encode('latin-1'),
                                     "Error in Nexus Python script", 0)
    sys.exit()


def messagebox(message, title=None):
    """ Custom notification handler """
    if title is None:
        title = "Message from Nexus Python script"
    ctypes.windll.user32.MessageBoxA(0, str(message), title, 0)


def yesno_box(message):
    """ Yes/no dialog with message """
    return ctypes.windll.user32.MessageBoxA(0,
                                            str(message), "Question", 1) == 1



# -*- coding: utf-8 -*-
"""
Normal data readers

@author: Jussi (jnu@iki.fi)
"""


from builtins import zip
import numpy as np
import openpyxl
import os.path as op
import logging

from .numutils import isfloat
from .models import models_all
from .config import cfg

logger = logging.getLogger(__name__)


def read_all_normaldata(age=None):
    """ Read all normal data defined in config. If age is specified, include
    age specific normaldata. """
    model_normaldata = dict()
    for fn in cfg.general.normaldata_files:
        ndata = read_normaldata(fn)
        model_normaldata.update(ndata)
    if age is not None:
        age_ndata_file = normaldata_age(age)
        if age_ndata_file:
            age_ndata = read_normaldata(age_ndata_file)
            model_normaldata.update(age_ndata)
    return model_normaldata


def read_normaldata(filename):
    """ Read normal data into dict. Dict keys are variables and values
    are Numpy arrays of shape (n, 2). n is either 1 (scalar variable)
    or 51 (data on 0..100% gait cycle, defined every 2% of cycle).
    The first and second columns are min and max values, respectively.
    (May be e.g. mean-stddev and mean+stddev)
    """
    if not op.isfile(filename):
        raise ValueError('No such file %s' % filename)
    type_ = op.splitext(filename)[1].lower()
    if type_ == '.gcd':
        return _read_gcd(filename)
    elif type_ == '.xlsx':
        return _read_xlsx(filename)
    else:
        raise ValueError('Only .gcd or .xlsx file formats are supported')


def normaldata_age(age):
    """ Return age specific normal data file """
    for age_range, filename in cfg.general.normaldata_age.items():
        if age_range[0] <= age <= age_range[1]:
            logger.debug('found normaldata file %s for age %d' %
                         (filename, age))
            return filename
    return None  # no matches


def _check_normaldata(ndata):
    """ Sanity checks """
    for val in ndata.values():
        if not all(np.diff(val) >= 0):
            raise ValueError('Normal data not in min/max format')
        if val.shape[0] not in [1, 51]:  # must be gait cycle data or scalar
            raise ValueError('Normal data has unexpected dimensions')
    return ndata


def _read_gcd(filename):
    """ Read normal data from a gcd file.
        -gcd data is assumed to be in (mean, dev) 2-column format and is
         converted to (min, max) (Polygon normal data format) as
         mean-dev, mean+dev
        -gcd variable names may be weird and will be translated according
        to each models translation table """
    ndata = dict()
    with open(filename, 'r') as f:
        lines = f.readlines()
    varname = None
    for li in lines:
        lis = li.split()
        if li[0] == '!':  # new variable
            varname = lis[0][1:]
            ndata[varname] = list()
        elif varname and isfloat(lis[0]):  # actual data
            # assume mean, dev format
            mean, dev = np.array(lis, dtype=float)
            ndata[varname].append([mean-dev, mean+dev])
        else:  # comment etc.
            continue
    # translate variable names
    ndata_ = dict()
    for nvarname, nval in ndata.items():
        for model in models_all:
            if nvarname in model.gcd_normaldata_map:
                logger.debug('mapping normal data variable %s -> %s' %
                             (nvarname, model.gcd_normaldata_map[nvarname]))
                nvarname = model.gcd_normaldata_map[nvarname]
                break
        ndata_[nvarname] = nval
    normaldata = {key: np.array(val) for key, val in ndata_.items()}
    return _check_normaldata(normaldata)


def _read_xlsx(filename):
    """ Read normal data exported from Polygon (xlsx format). """
    wb = openpyxl.load_workbook(filename)
    ws = wb['Normal']
    colnames = (cell.value for cell in next(ws.rows))  # first row: col names
    normaldata = dict()
    # read the columns and produce dict of numpy arrays
    for colname, col in zip(colnames, ws.columns):
        if colname is None:
            continue
        # pick values from row 4 onwards (skips units etc.)
        data = np.fromiter((c.value for k, c in enumerate(col) if k >= 3),
                           float)
        data = data[~np.isnan(data)]  # drop empty rows
        # rewrite the coordinate
        colname = colname.replace(' (1)', 'X')
        colname = colname.replace(' (2)', 'Y')
        colname = colname.replace(' (3)', 'Z')
        # scalar power variables (ones ending in 'Power')
        # are written as Z component (Nexus convention)
        if colname[-5:] == 'Power':
            colname += 'Z'
        normaldata[colname] = (np.stack([normaldata[colname], data], axis=1)
                               if colname in normaldata else data)
    return _check_normaldata(normaldata)



# -*- coding: utf-8 -*-
"""
Created on Tue Mar 17 14:41:31 2015

Computations on gait trials

@author: Jussi (jnu@iki.fi)
"""


from __future__ import division
from builtins import zip
import numpy as np
import logging

from .trial import Trial
from . import c3d

logger = logging.getLogger(__name__)


def get_analysis(c3dfile, condition='unknown'):
    """ A wrapper that reads the c3d analysis values """
    di = c3d.get_analysis(c3dfile, condition=condition)
    # Nexus <2.8 does not compute step width into c3d
    if 'Step Width' not in di[condition]:
        logger.warning('computing step widths (not found in %s)' % c3dfile)
        sw = _step_width(c3dfile)
        di[condition]['Step Width'] = dict()
        # XXX: uses avg of all cycles from trial
        di[condition]['Step Width']['Right'] = np.array(sw['R']).mean()
        di[condition]['Step Width']['Left'] = np.array(sw['L']).mean()
        di[condition]['Step Width']['unit'] = 'm'
    return di


def group_analysis(an_list, fun=np.mean):
    """ Average (or stddev etc) analysis dicts by applying fun to
    collected values. The condition label needs to be the same for all dicts.
    Returns single dict with the same condition. """

    if not isinstance(an_list, list):
        raise ValueError('Need a list of analysis dicts')

    if not an_list:
        return None

    if len(an_list) == 1:
        return an_list[0]

    condsets = [set(an.keys()) for an in an_list]
    if set.difference(*condsets):  # currently we are strict
        raise ValueError('Conditions need to match')
    conds = set.intersection(*condsets)

    for cond in conds:
        varsets = [set(an[cond].keys()) for an in an_list for cond in conds]

    vars_ = set.intersection(*varsets)
    not_in_all = set.union(*varsets) - vars_
    if not_in_all:
        logger.warning('Some files are missing the following variables: %s'
                       % ' '.join(not_in_all))
    res = dict()
    for cond in conds:
        res[cond] = dict()
        for var in vars_:
            res[cond][var] = dict()
            # this will fail if vars are not strictly matched between dicts
            res[cond][var]['unit'] = an_list[0][cond][var]['unit']
            for context in ['Right', 'Left']:
                # gather valus from analysis dicts
                allvals = np.array([an[cond][var][context] for an in an_list if
                                    context in an[cond][var]])
                # filter out missing values (nans)
                allvals = allvals[~np.isnan(allvals)]
                res[cond][var][context] = (fun(allvals) if allvals.size else
                                           np.nan)
    return res


def _step_width(source):
    """ Compute step width over trial cycles. See:
    https://www.vicon.com/faqs/software/how-does-nexus-plug-in-gait-and-polygon-calculate-gait-cycle-parameters-spatial-and-temporal
    Returns context keyed dict of lists.
    FIXME: marker name into params?
    FIXME: this (and similar) may also need to take Trial instance as argument
    to avoid creating new Trials
    """
    tr = Trial(source)
    sw = dict()
    mkr = 'TOE'  # marker name without context
    mkrdata = tr.full_marker_data
    # FIXME: why not use cycles here?
    for context, strikes in zip(['L', 'R'], [tr.lstrikes, tr.rstrikes]):
        sw[context] = list()
        nstrikes = len(strikes)
        if nstrikes < 2:
            continue
        # contralateral vars
        context_co = 'L' if context == 'R' else 'R'
        strikes_co = tr.lstrikes if context == 'R' else tr.rstrikes
        mname = context + mkr
        mname_co = context_co + mkr
        for j, strike in enumerate(strikes):
            if strike == strikes[-1]:  # last strike on this side
                break
            pos_this = mkrdata[mname][strike]
            pos_next = mkrdata[mname][strikes[j+1]]
            strikes_next_co = [k for k in strikes_co if k > strike]
            if len(strikes_next_co) == 0:  # no subsequent contralateral strike
                break
            pos_next_co = mkrdata[mname_co][strikes_next_co[0]]
            # vector distance between 'step lines' (see url above)
            V1 = pos_next - pos_this
            V1 /= np.linalg.norm(V1)
            VC = pos_next_co - pos_this
            VCP = V1 * np.dot(VC, V1)  # proj to ipsilateral line
            VSW = VCP - VC
            # marker data is in mm, but return step width in m
            sw[context].append(np.linalg.norm(VSW) / 1000.)
    return sw



# -*- coding: utf-8 -*-
"""
__init__.py

@author: Jussi (jnu@iki.fi)
"""
from __future__ import absolute_import

from .config import cfg
from . import models
from . import nexus
from . import eclipse
from . import read_data
from . import trial
from . import guiutils
from . import utils
from . import stats
from . import analysis
from . import plot_plotly
from . import sessionutils
from .emg import EMG
from .envutils import register_gui_exception_handler, GaitDataError
from .numutils import rising_zerocross, falling_zerocross
from .guiutils import messagebox
from .plot import Plotter
from .trial import Trial
from . import report



# -*- coding: utf-8 -*-
"""

Data readers & processing utils, Nexus specific

@author: Jussi (jnu@iki.fi)
"""

from __future__ import print_function, division
from collections import defaultdict
import sys
import numpy as np
import os.path as op
import psutil
import glob
import logging

from .numutils import change_coords
from .envutils import GaitDataError

from .config import cfg


logger = logging.getLogger(__name__)


def _find_nexus_path(vicon_path):
    """Return path to most recent Nexus version"""

    if vicon_path is None:
        vicon_path = r'C:\Program Files (x86)\Vicon'  # educated guess

    if not op.isdir(vicon_path):
        return None

    nexus_glob = op.join(vicon_path, 'Nexus?.*')
    nexus_dirs = glob.glob(nexus_glob)
    if not nexus_dirs:
        return None
    nexus_vers = [op.split(dir_)[1][5:] for dir_ in nexus_dirs]
    idx = nexus_vers.index(max(nexus_vers))
    return nexus_dirs[idx]


def _add_nexus_path(vicon_path):
    """Add Nexus SDK dir to sys.path"""

    nexus_path = _find_nexus_path(vicon_path)
    if nexus_path is None:
        print('cannot locate Nexus installation directory under %s'
              % vicon_path)
        return

    sdk_path = op.join(nexus_path, 'SDK', 'Python')
    if sdk_path not in sys.path:
        sys.path.append(sdk_path)
    else:
        print('%s already in sys.path' % sdk_path)

    # import from Win32 or Win64 according to bitness of Python interpreter
    bitness = '64' if sys.maxsize > 2**32 else '32'
    win = 'Win' + bitness
    _win_sdk_path = op.join(nexus_path, 'SDK', win)

    # check that the path for the wrong architecture has not already been
    # added to path (this may happen when running inside Nexus)
    win_other = 'Win32' if win == 'Win64' else 'Win64'
    _win_sdk_other = op.join(nexus_path, 'SDK', win_other)
    if _win_sdk_other in sys.path:
        print('%s already in sys.path, removing' % _win_sdk_other)
        sys.path.remove(_win_sdk_other)

    if _win_sdk_path not in sys.path:
        print('using Nexus SDK from %s' % _win_sdk_path)
        sys.path.append(_win_sdk_path)
    else:
        print('%s already in sys.path' % _win_sdk_path)


if sys.version_info.major >= 3:
    print('running on Python 3 or newer, cannot import Nexus API (yet)')
else:
    vicon_path = op.normpath(cfg.general.vicon_path)
    _add_nexus_path(vicon_path)
    try:
        import ViconNexus
    except ImportError:
        print('cannot import Vicon Nexus SDK')

sys.stdout.flush()  # make sure import warnings get printed


def pid():
    """ Tries to return the PID of the running Nexus process. """
    PROCNAME = "Nexus.exe"
    for proc in psutil.process_iter():
        try:
            if proc.name() == PROCNAME:
                return proc.pid
        # catch NoSuchProcess for procs that disappear inside loop
        except (psutil.AccessDenied, psutil.NoSuchProcess):
            pass
    return None


def true_ver():
    """ Tries to return the actual version of the running Nexus process
    (API does not do that). Hackish and probably unreliable """
    PROCNAME = "Nexus.exe"
    for proc in psutil.process_iter():
        try:
            if proc.name() == PROCNAME:
                exname = proc.exe()
                vind = exname.find('2.')  # assumes ver >2.
                if vind == -1:
                    return None
                try:
                    return float(exname[vind:vind+3])
                except ValueError:
                    return None
        except psutil.AccessDenied:
            pass
    return None


def viconnexus():
    """ Return a ViconNexus instance. """
    check_nexus()
    return ViconNexus.ViconNexus()


def get_subjectnames(single_only=True):
    """ Get subject name(s) from Nexus """
    vicon = viconnexus()
    get_sessionpath()  # check whether we can get data
    names_ = vicon.GetSubjectNames()
    if not names_:
        raise GaitDataError('No subject defined in Nexus')
    if single_only:
        if len(names_) > 1:
            raise GaitDataError('Nexus returns multiple subjects')
    """ Workaround a Nexus 2.6 bug (?) that creates extra names with
    weird unicode strings """
    names_ = [name for name in names_ if u'\ufffd1' not in name]
    return names_[0] if single_only else names_


def check_nexus():
    if not pid():
        raise GaitDataError('Vicon Nexus does not seem to be running')


def get_sessionpath():
    """ Get path to current session """
    try:
        vicon = viconnexus()
        sessionpath = vicon.GetTrialName()[0]
    except IOError:  # may be raised if Nexus was just terminated
        sessionpath = None
    if not sessionpath:
        raise GaitDataError('Cannot get Nexus session path, '
                            'no session or maybe in Live mode?')
    return op.normpath(sessionpath)


def run_pipelines(vicon, plines):
    """Run given Nexus pipeline(s)"""
    if type(plines) != list:
        plines = [plines]
    for pipeline in plines:
        logger.debug('running pipeline: %s' % pipeline)
        result = vicon.Client.RunPipeline(pipeline.encode('utf-8'), '',
                                          cfg.autoproc.nexus_timeout)
        if result.Error():
            logger.warning('error while trying to run Nexus pipeline: %s'
                           % pipeline)


def get_trialname():
    """ Get trial name without session path """
    vicon = viconnexus()
    trialname_ = vicon.GetTrialName()
    return trialname_[1]


def is_vicon_instance(obj):
    """ Check if obj is an instance of ViconNexus """
    return obj.__class__.__name__ == 'ViconNexus'


def _get_nexus_subject_param(vicon, name, param):
    """Get subject parameter from Nexus."""
    value = vicon.GetSubjectParam(name, param)
    # for unknown reasons, above method may return tuple or float
    # depending on whether script is run from Nexus or outside
    if type(value) == tuple:
        value = value[0] if value[1] else None
    return value


def _get_marker_names(vicon, trajs_only=True):
    """Return marker names (only ones with trajectories, if trajs_only)"""
    subjname = get_subjectnames()
    markers = vicon.GetMarkerNames(subjname)
    # only get markers with trajectories - excludes calibration markers
    if trajs_only:
        markers = [mkr for mkr in markers if vicon.HasTrajectory(subjname,
                                                                 mkr)]
    return markers


# FIXME: most of get_ methods below are intended to be called via read_data
# so underscore them
def get_metadata(vicon):
    """ Read trial and subject metadata """
    check_nexus()
    logger.debug('reading metadata from Vicon Nexus')
    subjname = get_subjectnames()
    params_available = vicon.GetSubjectParamNames(subjname)
    subj_params = defaultdict(lambda: None)
    subj_params.update({par: _get_nexus_subject_param(vicon, subjname, par) for
                        par in params_available})
    trialname = get_trialname()
    if not trialname:
        raise GaitDataError('No trial loaded in Nexus')
    sessionpath = get_sessionpath()
    markers = _get_marker_names(vicon)
    # get events - GetEvents() indices seem to often be 1 frame less than on
    # Nexus display - only happens with ROI?
    lstrikes = vicon.GetEvents(subjname, "Left", "Foot Strike")[0]
    rstrikes = vicon.GetEvents(subjname, "Right", "Foot Strike")[0]
    ltoeoffs = vicon.GetEvents(subjname, "Left", "Foot Off")[0]
    rtoeoffs = vicon.GetEvents(subjname, "Right", "Foot Off")[0]
    # Offset will be subtracted from event frame numbers to get correct
    # 0-based index for frame data. For Nexus, it is always 1 (Nexus uses
    # 1-based frame numbering)
    offset = 1
    length = vicon.GetFrameCount()
    framerate = vicon.GetFrameRate()
    # Get analog rate. This may not be mandatory if analog devices
    # are not used, but currently it needs to succeed.
    devids = vicon.GetDeviceIDs()
    if not devids:
        raise GaitDataError('Cannot determine analog rate')
    else:
        devid = devids[0]
        _, _, analograte, _, _, _ = vicon.GetDeviceDetails(devid)
    samplesperframe = analograte / framerate
    logger.debug('offset @ %d, %d frames, framerate %d Hz, %d samples per '
                 'frame' % (offset, length, framerate, samplesperframe))
    # get n of forceplates
    fp_devids = [id_ for id_ in devids if
                 vicon.GetDeviceDetails(id_)[1].lower() == 'forceplate']

    # sort events (may be in wrong temporal order, at least in c3d files)
    for li in [lstrikes, rstrikes, ltoeoffs, rtoeoffs]:
        li.sort()

    return {'trialname': trialname, 'sessionpath': sessionpath,
            'offset': offset, 'framerate': framerate, 'analograte': analograte,
            'name': subjname, 'subj_params': subj_params, 'lstrikes': lstrikes,
            'rstrikes': rstrikes, 'ltoeoffs': ltoeoffs, 'rtoeoffs': rtoeoffs,
            'length': length, 'samplesperframe': samplesperframe,
            'n_forceplates': len(fp_devids), 'markers': markers}


def get_emg_data(vicon):
    """ Read EMG data from Nexus. This uses the configured EMG device name. """
    return _get_analog_data(vicon, cfg.emg.devname)


def get_accelerometer_data(vicon):
    """ Read EMG data from Nexus. This uses the configured EMG device name. """
    return _get_analog_data(vicon, cfg.analog.accelerometer_devname)


def _get_analog_data(vicon, devname):
    """ Read analog data from Nexus """
    ids = [id_ for id_ in vicon.GetDeviceIDs() if
           vicon.GetDeviceDetails(id_)[0].lower() == devname.lower()]
    if len(ids) > 1:
        raise GaitDataError('Multiple matching analog devices')
    elif len(ids) == 0:
        raise GaitDataError('No matching analog devices')
    dev_id = ids[0]
    dname, dtype, drate, outputids, _, _ = vicon.GetDeviceDetails(dev_id)
    # not handling multiple output ids yet
    if len(outputids) != 1:
        raise GaitDataError('Expected single output for device')
    outputid = outputids[0]
    # get list of channel names and IDs
    _, _, _, _, chnames, chids = vicon.GetDeviceOutputDetails(dev_id, outputid)
    data = dict()
    for chid in chids:
        chdata, _, chrate = vicon.GetDeviceChannel(dev_id, outputid, chid)
        chname = chnames[chid-1]  # chids start from 1
        data[chname] = np.array(chdata)
    return {'t': np.arange(len(chdata)) / drate, 'data': data}


def _get_1_forceplate_data(vicon, devid):
    """ Read data of single forceplate from Nexus.
    Data is returned in global coordinate frame """
    # get forceplate ids
    logger.debug('reading forceplate data from devid %d' % devid)
    dname, dtype, drate, outputids, nfp, _ = vicon.GetDeviceDetails(devid)
    # outputs should be force, moment, cop. select force
    outputid = outputids[0]
    chid = vicon.GetDeviceChannelIDFromName(devid, outputid, 'Fx')
    fx, chready, chrate = vicon.GetDeviceChannelGlobal(devid, outputid, chid)
    chid = vicon.GetDeviceChannelIDFromName(devid, outputid, 'Fy')
    fy, chready, chrate = vicon.GetDeviceChannelGlobal(devid, outputid, chid)
    chid = vicon.GetDeviceChannelIDFromName(devid, outputid, 'Fz')
    fz, chready, chrate = vicon.GetDeviceChannelGlobal(devid, outputid, chid)
    # moments
    outputid = outputids[1]
    chid = vicon.GetDeviceChannelIDFromName(devid, outputid, 'Mx')
    mx, chready, chrate = vicon.GetDeviceChannelGlobal(devid, outputid, chid)
    chid = vicon.GetDeviceChannelIDFromName(devid, outputid, 'My')
    my, chready, chrate = vicon.GetDeviceChannelGlobal(devid, outputid, chid)
    chid = vicon.GetDeviceChannelIDFromName(devid, outputid, 'Mz')
    mz, chready, chrate = vicon.GetDeviceChannelGlobal(devid, outputid, chid)
    # center of pressure
    outputid = outputids[2]
    chid = vicon.GetDeviceChannelIDFromName(devid, outputid, 'Cx')
    copx, chready, chrate = vicon.GetDeviceChannelGlobal(devid, outputid, chid)
    chid = vicon.GetDeviceChannelIDFromName(devid, outputid, 'Cy')
    copy, chready, chrate = vicon.GetDeviceChannelGlobal(devid, outputid, chid)
    chid = vicon.GetDeviceChannelIDFromName(devid, outputid, 'Cz')
    copz, chready, chrate = vicon.GetDeviceChannelGlobal(devid, outputid, chid)
    cop_w = np.array([copx, copy, copz]).transpose()
    F = np.array([fx, fy, fz]).transpose()
    M = np.array([mx, my, mz]).transpose()
    Ftot = np.linalg.norm(F, axis=1)
    # translation and rotation matrices -> world coords
    # suspect that Nexus wR is wrong (does not match displayed plate axes)?
    wR = np.array(nfp.WorldR).reshape(3, 3)
    wT = np.array(nfp.WorldT)
    # plate corners -> world coords
    cor = np.stack([nfp.LowerBounds, nfp.UpperBounds])
    cor_w = change_coords(cor, wR, wT)
    cor_full = np.array([cor_w[0, :],
                        [cor_w[0, 0], cor_w[1, 1], cor_w[0, 2]],
                        cor_w[1, :],
                        [cor_w[1, 0], cor_w[0, 1], cor_w[0, 2]]])

    lb = np.min(cor_w, axis=0)
    ub = np.max(cor_w, axis=0)
    # check that CoP stays inside plate boundaries
    cop_ok = np.logical_and(cop_w[:, 0] >= lb[0], cop_w[:, 0] <= ub[0]).all()
    cop_ok &= np.logical_and(cop_w[:, 1] >= lb[1], cop_w[:, 1] <= ub[1]).all()
    if not cop_ok:
        logger.warning('center of pressure outside plate boundaries, '
                       'clipping to plate')
        cop_w[:, 0] = np.clip(cop_w[:, 0], lb[0], ub[0])
        cop_w[:, 1] = np.clip(cop_w[:, 1], lb[1], ub[1])
    return {'F': F, 'M': M, 'Ftot': Ftot, 'CoP': cop_w,
            'wR': wR, 'wT': wT, 'lowerbounds': lb, 'upperbounds': ub,
            'cor_w': cor_w, 'cor_full': cor_full}


def get_forceplate_data(vicon):
    """ Read all forceplate data from Nexus. """
    # get forceplate ids
    logger.debug('reading forceplate data from Vicon Nexus')
    devids = [id_ for id_ in vicon.GetDeviceIDs() if
              vicon.GetDeviceDetails(id_)[1].lower() == 'forceplate']
    if len(devids) == 0:
        logger.debug('no forceplates detected')
        return None
    logger.debug('detected %d forceplate(s)' % len(devids))
    return [_get_1_forceplate_data(vicon, devid) for devid in devids]


def _swap_markers(vicon, marker1, marker2):
    """Swap two marker trajectories in currently loaded trial."""
    subj = get_subjectnames()
    m1 = vicon.GetTrajectory(subj, marker1)
    m2 = vicon.GetTrajectory(subj, marker2)
    vicon.SetTrajectory(subj, marker2, m1[0], m1[1], m1[2], m1[3])
    vicon.SetTrajectory(subj, marker1, m2[0], m2[1], m2[2], m2[3])


def _get_marker_data(vicon, markers, ignore_edge_gaps=True,
                     ignore_missing=False):
    """Get position, velocity and acceleration for specified markers."""
    if not isinstance(markers, list):
        markers = [markers]
    subj = get_subjectnames()
    mkrdata = dict()
    for marker in markers:
        x, y, z, _ = vicon.GetTrajectory(subj, marker)
        if len(x) == 0:
            if ignore_missing:
                logger.warning('Cannot read trajectory %s from Nexus'
                               % marker)
                continue
            else:
                raise GaitDataError('Cannot read marker trajectory '
                                    'from Nexus: \'%s\'' % marker)
        mP = np.array([x, y, z]).transpose()
        mkrdata[marker] = mP
        mkrdata[marker + '_P'] = mP
        mkrdata[marker + '_V'] = np.gradient(mP)[0]
        mkrdata[marker + '_A'] = np.gradient(mkrdata[marker+'_V'])[0]
        # find gaps
        allzero = np.any(mP, axis=1).astype(int)
        if ignore_edge_gaps:
            nleading = allzero.argmax()
            allzero_trim = np.trim_zeros(allzero)
            gap_inds = np.where(allzero_trim == 0)[0] + nleading
        else:
            gap_inds = np.where(allzero == 0)[0]
        mkrdata[marker + '_gaps'] = gap_inds
    return mkrdata


def get_roi(vicon):
    """ Return array of frames corresponding to Nexus ROI """
    roi = vicon.GetTrialRegionOfInterest()
    return np.arange(roi[0], roi[1])


def get_model_data(vicon, model):
    """ Read model output variables (e.g. Plug-in Gait) """
    modeldata = dict()
    var_dims = (3, vicon.GetFrameCount())
    subj = get_subjectnames()
    for var in model.read_vars:
        nums, bools = vicon.GetModelOutput(subj, var)
        if nums:
            data = np.squeeze(np.array(nums))
        else:
            if model.is_optional_var(var):
                logger.info('cannot read optional variable %s, returning nans'
                            % var)
                data = np.empty(var_dims)
                data[:] = np.nan
            else:
                raise GaitDataError('Cannot read model variable %s. Make sure '
                                    'that the appropriate model has been run.'
                                    % var)
        modeldata[var] = data
    return modeldata


def create_events(vicon, context, strikes, toeoffs):
    """ Create foot strike and toeoff events in Nexus. """
    logger.debug('marking events in Nexus')
    side_str = 'Right' if context == 'R' else 'Left'
    subjectname = get_subjectnames()
    for fr in strikes:
        vicon.CreateAnEvent(subjectname, side_str, 'Foot Strike',
                            int(fr+1), 0)
    for fr in toeoffs:
        vicon.CreateAnEvent(subjectname, side_str, 'Foot Off',
                            int(fr+1), 0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
PyQt graphical interface to gaitutils

@author: Jussi (jnu@iki.fi)
"""

from __future__ import print_function
from builtins import str
from PyQt5 import QtGui, QtCore, uic, QtWidgets
from PyQt5.QtCore import QRunnable, QThreadPool, pyqtSignal, QObject
from pkg_resources import resource_filename
from functools import partial
import sys
import os.path as op
import os
import subprocess
import time
import requests
import logging
import traceback
import itertools

from .qt_dialogs import (OptionsDialog, qt_message_dialog, qt_yesno_dialog,
                         ChooseSessionsDialog)
from .qt_widgets import QtHandler, ProgressBar, ProgressSignals, XStream
from ..numutils import check_hetu
from .. import (GaitDataError, nexus, cfg, report, sessionutils, videos,
                envutils)
from . import _tardieu
from ..scripts import (nexus_plot,
                       nexus_emg_consistency,
                       nexus_kin_consistency,
                       nexus_musclelen_consistency,
                       nexus_autoprocess_trial,
                       nexus_autoprocess_session,
                       nexus_copy_trial_videos,
                       nexus_trials_velocity,
                       nexus_make_pdf_report,
                       nexus_make_comparison_report,
                       nexus_kin_average,
                       nexus_automark_trial,
                       nexus_time_distance_vars)


logger = logging.getLogger(__name__)


def _exception_msg(e):
    """Return text representation of exception e"""
    # for our own error class, we know that a neat message is there
    if isinstance(e, GaitDataError):
        err_msg = e.message
    else:  # otherwise, we have no idea, so use generic repr()
        err_msg = repr(e)
    return 'There was an error running the operation. Details:\n%s' % err_msg


def _collect_videos_to_convert(session, tags):
    """Collect session AVI files for conversion to web format"""
    c3ds = sessionutils.get_c3ds(session, tags=tags,
                                 trial_type='dynamic')
    c3ds += sessionutils.get_c3ds(session, tags=cfg.eclipse.video_tags,
                                  trial_type='dynamic')
    c3ds += sessionutils.get_c3ds(session, trial_type='static')
    vids_it = (videos.get_trial_videos(c3d, vid_ext='.avi')
               for c3d in c3ds)
    return list(itertools.chain.from_iterable(vids_it))


class PdfReportDialog(QtWidgets.QDialog):
    """Ask for patient/session info and report options"""

    def __init__(self, info, prompt='Hello', parent=None):
        super(self.__class__, self).__init__()
        uifile = resource_filename('gaitutils', 'gui/pdf_report_dialog.ui')
        uic.loadUi(uifile, self)
        #self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.prompt.setText(prompt)
        if info is not None:
            if info['fullname'] is not None:
                self.lnFullName.setText(info['fullname'])
            if info['hetu'] is not None:
                self.lnHetu.setText(info['hetu'])
            if info['session_description'] is not None:
                self.lnDescription.setText(info['session_description'])

    def accept(self):
        """ Update config and close dialog, if widget inputs are ok. Otherwise
        show an error dialog """
        self.hetu = self.lnHetu.text()
        self.fullname = self.lnFullName.text()
        self.session_description = self.lnDescription.text()
        # get all the report page selections
        self.pages = dict()
        for w in self.findChildren(QtWidgets.QWidget):
            wname = w.objectName()
            if wname[:2] == 'cb':
                self.pages[wname[2:]] = w.checkState()
        if self.fullname and check_hetu(self.hetu):
            self.done(QtWidgets.QDialog.Accepted)  # or call superclass accept
        else:
            qt_message_dialog('Please enter a valid name and hetu')


class WebReportInfoDialog(QtWidgets.QDialog):
    """Ask for patient info"""

    def __init__(self, info, parent=None, check_info=True):
        super(self.__class__, self).__init__()
        uifile = resource_filename('gaitutils', 'gui/web_report_info.ui')
        uic.loadUi(uifile, self)
        #self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.check_info = check_info
        if info is not None:
            if info['fullname'] is not None:
                self.lnFullName.setText(info['fullname'])
            if info['hetu'] is not None:
                self.lnHetu.setText(info['hetu'])
            if info['report_notes'] is not None:
                self.txtNotes.setPlainText(info['report_notes'])

    def accept(self):
        """ Update config and close dialog, if widget inputs are ok. Otherwise
        show an error dialog """
        self.hetu = self.lnHetu.text().strip()
        self.fullname = self.lnFullName.text().strip()
        self.report_notes = str(self.txtNotes.toPlainText()).strip()
        if self.check_info:
            ok = self.fullname and check_hetu(self.hetu)
        else:
            ok = not self.hetu or check_hetu(self.hetu)
        if ok:
            self.done(QtWidgets.QDialog.Accepted)  # or call superclass accept
        else:
            msg = 'Please enter a valid name and hetu'
            if not self.check_info:
                msg += ' (or leave empty)'
            qt_message_dialog(msg)


class WebReportDialog(QtWidgets.QDialog):
    """Dialog for managing web reports. In current implementation, this needs a
    GaitMenu instance as a parent (uses _execute() and other parent methods)"""

    def __init__(self, parent):
        super(self.__class__, self).__init__(parent)
        self.parent = parent
        # load user interface made with designer
        uifile = resource_filename('gaitutils', 'gui/web_report_dialog.ui')
        uic.loadUi(uifile, self)
        self.btnCreateReport.clicked.connect(self._create_web_report)
        self.btnDeleteReport.clicked.connect(self._delete_current_report)
        self.btnDeleteAllReports.clicked.connect(self._delete_all_reports)
        self.btnViewReport.clicked.connect(self._view_current_report)
        # add double click action to browse current report
        (self.listActiveReports.itemDoubleClicked.
         connect(lambda item: self._browse_localhost(item.userdata)))
        # these require active reports to be enabled
        self.reportWidgets = [self.btnDeleteReport, self.btnDeleteAllReports,
                              self.btnViewReport]
        self._set_report_button_status()
        self._browser_procs = list()

    def _create_web_report(self):
        """Collect sessions, create the dash app, start it and launch a
        web browser on localhost on the correct port"""

        if self.listActiveReports.count() == cfg.web_report.max_reports:
            qt_message_dialog('Maximum number of active web reports active. '
                              'Please delete some reports first.')
            return

        dlg = ChooseSessionsDialog()
        if not dlg.exec_():
            return
        sessions = dlg.sessions

        sessions_str = '/'.join([op.split(s)[-1] for s in sessions])
        report_type = ('single session' if len(sessions) == 1
                       else 'comparison')
        report_name = '%s: %s' % (report_type, sessions_str)
        existing_names = [item.text for item in self.listActiveReports.items]
        if report_name in existing_names:
            qt_message_dialog('There is already a report for %s' %
                              report_name)
            return

        session_infos, info = sessionutils._merge_session_info(sessions)
        if info is None:
            qt_message_dialog('Patient files do not match. Sessions may be '
                              'from different patients. Continuing without '
                              'patient info.')
            info = sessionutils.default_info()
        else:
            dlg_info = WebReportInfoDialog(info, check_info=False)
            if dlg_info.exec_():
                new_info = dict(hetu=dlg_info.hetu, fullname=dlg_info.fullname,
                                report_notes=dlg_info.report_notes)
                info.update(new_info)

                # update info files (except session specific keys)
                for session in sessions:
                    update_dict = dict(report_notes=dlg_info.report_notes,
                                       fullname=dlg_info.fullname,
                                       hetu=dlg_info.hetu)
                    session_infos[session].update(update_dict)
                    sessionutils.save_info(session, session_infos[session])
            else:
                return

        prog = ProgressBar('Creating web report...')
        prog.update('Collecting session information...', 0)
        signals = ProgressSignals()
        signals.progress.connect(lambda text, p: prog.update(text, p))

        # for comparison between sessions, get representative trials only
        tags = (cfg.eclipse.repr_tags if len(sessions) > 1 else
                cfg.eclipse.tags)

        # collect all video files for conversion
        # includes tagged dynamic, video-only tagged, and static trials
        vidfiles = list()
        for session in sessions:
            vids = _collect_videos_to_convert(session, tags=tags)
            vidfiles.extend(vids)

        if not report.convert_videos(vidfiles, check_only=True):
            self.parent._convert_vidfiles(vidfiles, signals)

        self._report_creation_status = None
        self.parent._execute(report.dash_report, thread=True, block_ui=True,
                             finished_func=self._web_report_finished,
                             info=info, sessions=sessions, tags=tags,
                             signals=signals)

        # wait for report creation thread to complete
        while self._report_creation_status is None:
            time.sleep(.05)
            QtWidgets.QApplication.processEvents()
        prog.close()

        if self._report_creation_status is False:
            qt_message_dialog('Could not create report, check that session is '
                              'valid')
            return
        app = self._report_creation_status

        # figure out first free TCP port
        ports_taken = [item.userdata for item in self.listActiveReports.items]
        port = cfg.web_report.tcp_port
        while port in ports_taken:  # find first port not taken by us
            port += 1

        # report ok - start server in a thread
        # also enable the threaded mode of the server. serving is a bit flaky
        # in Python 2 (multiple requests cause exceptions)
        self.parent._execute(app.server.run, thread=True, block_ui=False,
                             debug=False, port=port, threaded=True)
        # double clicking on the list item will browse to corresponding port
        self.listActiveReports.add_item(report_name, data=port)
        # enable delete buttons etc.
        self._set_report_button_status()
        logger.debug('starting web browser')
        self._browse_localhost(port)

    @property
    def active_reports(self):
        """Return number of active web reports"""
        return self.listActiveReports.count()

    def shutdown(self):
        """Try to shutdown browser processes and web servers"""
        for proc in self._browser_procs:
            proc.kill()
        # cannot use generator here since the loop changes the items
        for item in list(self.listActiveReports.items):
            self._delete_report(item)

    def _delete_report(self, item):
        """Shut down server for given list item, remove item"""
        port = item.userdata
        # compose url for shutdown request - see report.py
        url = 'http://127.0.0.1:%d/shutdown' % port
        # we have to make sure that localhost is not proxied
        proxies = {"http": None, "https": None}
        logger.debug('requesting server shutdown for port %d' % port)
        requests.get(url, proxies=proxies)
        self.listActiveReports.rm_current_item()

    def _delete_current_report(self):
        """Shut down server for current item, remove item"""
        item = self.listActiveReports.currentItem()
        if item is None:
            return
        msg = 'Are you sure you want to delete the report for %s?' % item.text
        reply = qt_yesno_dialog(msg)
        if reply == QtWidgets.QMessageBox.YesRole:
            self._delete_report(item)
        self._set_report_button_status()

    def _delete_all_reports(self):
        """Delete all web reports"""
        if self.listActiveReports.count() == 0:
            return
        msg = 'Are you sure you want to delete all reports?'
        reply = qt_yesno_dialog(msg)
        if reply != QtWidgets.QMessageBox.YesRole:
            return
        # cannot use generator here since the loop changes the items
        for item in list(self.listActiveReports.items):
            self._delete_report(item)
        self._set_report_button_status()

    def _view_current_report(self):
        """Open current report in browser"""
        item = self.listActiveReports.currentItem()
        if item is None:
            return
        port = item.userdata
        self._browse_localhost(port)

    def _set_report_button_status(self):
        """Enable report buttons if reports exist, otherwise disable them"""
        n_reports = self.active_reports
        for widget in self.reportWidgets:
            widget.setEnabled(True if n_reports else False)

    def _web_report_finished(self, app):
        """Gets called when web report creation is finished"""
        logger.debug('report creation finished')
        self._report_creation_status = app
        # this enables controls on all windows
        self.parent._enable_op_buttons(None)

    def _browse_localhost(self, port):
        """Open configured browser on localhost:port"""
        url = '127.0.0.1:%d' % port
        try:
            proc = subprocess.Popen([cfg.general.browser_path, url])
            self._browser_procs.append(proc)
            logger.debug('new browser pid %d' % proc.pid)
        except Exception:
            qt_message_dialog('Cannot start configured web browser: %s'
                              % cfg.general.browser_path)


class Gaitmenu(QtWidgets.QMainWindow):

    def __init__(self):
        super(self.__class__, self).__init__()
        # load user interface made with designer
        uifile = resource_filename('gaitutils', 'gui/gaitmenu.ui')
        uic.loadUi(uifile, self)
        """
        matplotlib and threads:
        -----------------------
        Stuff that shows matplotlib plots cannot be run in directly in
        worker threads. To put plotting stuff into a worker thread, we need to:
        -make the plotting function return a figure (and not invoke the qt
        event loop)
        -put the plotting function into a worker thread
        -call plotting function
        -create a qt dialog and canvas in GUI thread
        -draw resulting figure onto canvas
        So far this has not been done since the plotting functions return
        rather quickly.
        Running the plotting functions directly from the GUI thread is also a
        bit ugly since the Qt event loop gets called twice, but this does not
        seem to do any harm.

        launching web servers:
        ----------------------
        Web servers need to go into separate threads/processes so that the rest
        of the app can continue running. It's hard to use processes because of
        problems with multiprocessing/pickle, so the Qt threadpool is used to
        launch the servers. However since each running server occupies a
        thread, this means that we need to increase the threadpool max threads
        limit; otherwise new servers will get queued by the threadpool and will
        not run.
        """

        if not cfg.general.allow_multiple_menu_instances:
            sname = op.split(__file__)[1]
            nprocs = envutils._count_script_instances(sname)
            if nprocs >= 2:
                qt_message_dialog('Another instance of the menu seems to be '
                                  'running. Please use that instance or '
                                  'stop it before starting a new one.')
                sys.exit()

        if cfg.general.git_autoupdate:
            if envutils._git_autoupdate():
                qt_message_dialog('The package was automatically updated. '
                                  'Restarting...')
                os.execv(sys.executable, ['python'] + sys.argv)

        # if using the plotly backend, we can run plotters in worker threads
        thread_plotters = cfg.plot.backend == 'plotly'

        self._web_report_dialog = WebReportDialog(self)

        # modal dialogs etc. (simple signal->slot connection)
        self.actionCreate_PDF_report.triggered.connect(self._create_pdf_report)
        self.actionWeb_reports.triggered.connect(self._web_report_dialog.show)
        self.actionQuit.triggered.connect(self.close)
        self.actionOpts.triggered.connect(self._options_dialog)
        self.actionTardieu_analysis.triggered.connect(self._tardieu)
        self.actionAutoprocess_session.triggered.connect(self._autoproc_session)
        self.btnPlotNexusTrial.clicked.connect(self._plot_nexus_trial)

        # consistency menu
        self._widget_connect_task(self.actionTrial_velocity,
                                  nexus_trials_velocity.do_plot)
        self._widget_connect_task(self.actionTime_distance_average,
                                  nexus_time_distance_vars.
                                  do_session_average_plot)
        self._widget_connect_task(self.actionKin_average,
                                  nexus_kin_average.do_plot)
        self._widget_connect_task(self.actionKinematics_consistency,
                                  nexus_kin_consistency.do_plot,
                                  thread=thread_plotters)
        self._widget_connect_task(self.actionEMG_consistency,
                                  nexus_emg_consistency.do_plot,
                                  thread=thread_plotters)
        self._widget_connect_task(self.actionMuscle_length_consistency,
                                  nexus_musclelen_consistency.do_plot,
                                  thread=thread_plotters)

        # processing menu
        self._widget_connect_task(self.actionAutoprocess_single_trial,
                                  nexus_autoprocess_trial.autoproc_single,
                                  thread=True)
        self._widget_connect_task(self.actionAutomark_events,
                                  nexus_automark_trial.automark_single,
                                  thread=True)
        self._widget_connect_task(self.actionRun_postprocessing_pipelines,
                                  self._postprocess_session)
        self._widget_connect_task(self.actionConvert_session_videos_to_web_format,
                                  self._convert_session_videos)
        self._widget_connect_task(self.actionCopy_session_videos_to_desktop,
                                  nexus_copy_trial_videos.do_copy)

        # add predefined plot layouts to combobox
        cb_items = sorted(cfg.layouts.menu_layouts.keys())
        self.cbNexusTrialLayout.addItems(cb_items)
        # set default option to PiG lower body (if it's on the list)
        try:
            default_index = cb_items.index('PiG lower body kinematics+'
                                           'kinetics')
        except ValueError:
            default_index = 0
        self.cbNexusTrialLayout.setCurrentIndex(default_index)

        XStream.stdout().messageWritten.connect(self._log_message)
        XStream.stderr().messageWritten.connect(self._log_message)
        logger.debug('interpreter: %s' % sys.executable)
        self.threadpool = QThreadPool()
        # we need a thread for each web server plus one worker thread
        self.threadpool.setMaxThreadCount(cfg.web_report.max_reports + 1)
        self._tardieuwin = None

    def _autoproc_session(self):
        """Wrapper to run autoprocess for Nexus session"""
        try:
            sessionpath = nexus.get_sessionpath()
        except GaitDataError as e:
            self._exception(e)
            return
        c3ds = sessionutils.get_c3ds(sessionpath, trial_type='DYNAMIC',
                                     check_if_exists=True)
        if c3ds:
            reply = qt_yesno_dialog('Some of the dynamic trials have been '
                                    'processed already. Are you sure you want '
                                    'to run autoprocessing?')
            if reply == QtWidgets.QMessageBox.NoRole:
                return

        self._execute(nexus_autoprocess_session.autoproc_session,
                      thread=True,
                      finished_func=self._enable_op_buttons)

    def _plot_nexus_trial(self):
        """Plot the current Nexus trial"""
        lout_desc = self.cbNexusTrialLayout.currentText()
        lout_name = cfg.layouts.menu_layouts[lout_desc]
        cycs = 'unnormalized' if self.xbPlotUnnorm.checkState() else None
        model_cycles = emg_cycles = cycs
        from_c3d = self.xbPlotFromC3D.checkState()
        self._execute(nexus_plot.do_plot, thread=True,
                      finished_func=self._enable_op_buttons,
                      layout_name=lout_name, model_cycles=model_cycles,
                      emg_cycles=emg_cycles, from_c3d=from_c3d)

    def _widget_connect_task(self, widget, fun, thread=False):
        """Helper to connect widget with task. Use lambda to consume unused
        events argument. If thread=True, launch in a separate worker thread."""
        # by default, just enable UI buttons when thread finishes
        finished_func = self._enable_op_buttons if thread else None
        if isinstance(widget, QtWidgets.QPushButton):
            sig = widget.clicked
        elif isinstance(widget, QtWidgets.QAction):
            sig = widget.triggered
        else:
            raise ValueError('Invalid widget type')
        sig.connect(lambda ev: self._execute(fun, thread=thread,
                                             finished_func=finished_func))

    def _convert_vidfiles(self, vidfiles, signals):
        """Convert given list of video files to web format. Uses non-blocking
        Popen() calls"""
        self._disable_op_buttons()
        procs = self._execute(report.convert_videos, thread=False,
                              block_ui=False, vidfiles=vidfiles)
        if not procs:
            return

        completed = False
        while not completed:
            n_complete = len([p for p in procs if p.poll() is not None])
            prog_txt = ('Converting videos: %d of %d files done'
                        % (n_complete, len(procs)))
            prog_p = 100 * n_complete / float(len(procs))
            signals.progress.emit(prog_txt, prog_p)
            time.sleep(.25)
            completed = n_complete == len(procs)
        self._enable_op_buttons(None)

    def _convert_session_videos(self):
        """Convert current Nexus session videos to web format."""
        try:
            session = nexus.get_sessionpath()
        except GaitDataError as e:
            qt_message_dialog(_exception_msg(e))
            return
        try:
            vidfiles = _collect_videos_to_convert(session,
                                                  tags=cfg.eclipse.tags)
        except GaitDataError as e:
            qt_message_dialog(_exception_msg(e))
            return
        if not vidfiles:
            qt_message_dialog('Cannot find any video files for session %s'
                              % session)
            return
        if report.convert_videos(vidfiles, check_only=True):
            qt_message_dialog('It looks like the session videos have already '
                              'been converted.')
            return
        prog = ProgressBar('')
        signals = ProgressSignals()
        signals.progress.connect(lambda text, p: prog.update(text, p))
        self._convert_vidfiles(vidfiles, signals)
        prog.close()

    def _postprocess_session(self):
        """Run additional postprocessing pipelines for tagged trials"""
        try:
            session = nexus.get_sessionpath()
        except GaitDataError as e:
            qt_message_dialog(_exception_msg(e))
            return
        # XXX: run for tagged + static - maybe this should be configurable
        trials = sessionutils.get_c3ds(session, tags=cfg.eclipse.tags,
                                       trial_type='dynamic')
        trials += sessionutils.get_c3ds(session, trial_type='static')
        if trials and cfg.autoproc.postproc_pipelines:
            logger.debug('running postprocessing for %s' % trials)
            prog = ProgressBar('')
            vicon = nexus.viconnexus()
            prog.update('Running postprocessing pipelines: %s for %d '
                        'trials' % (cfg.autoproc.postproc_pipelines,
                                    len(trials)), 0)
            for k, tr in enumerate(trials):
                trbase = op.splitext(tr)[0]
                vicon.OpenTrial(trbase, cfg.autoproc.nexus_timeout)
                nexus.run_pipelines(vicon, cfg.autoproc.postproc_pipelines)
                prog.update('Running postprocessing pipelines: %s for %d '
                            'trials' % (cfg.autoproc.postproc_pipelines, len(trials)), 100*k/len(trials))
        elif not trials:
            qt_message_dialog('No trials in session to run postprocessing for')

    def closeEvent(self, event):
        """ Confirm and close application. """
        if self._web_report_dialog.active_reports:
            reply = qt_yesno_dialog('There are active web reports which '
                                    'will be closed. Are you sure you '
                                    'want to quit?')
            if reply == QtWidgets.QMessageBox.YesRole:
                self._web_report_dialog.shutdown()
                event.accept()
            else:
                event.ignore()
        else:
            event.accept()

    def _options_dialog(self):
        """Show the options dialog"""
        dlg = OptionsDialog(self)
        dlg.exec_()

    def _create_pdf_report(self):
        """Create comparison or single session pdf report"""

        dlg = ChooseSessionsDialog()
        if not dlg.exec_():
            return
        sessions = dlg.sessions
        comparison = len(sessions) > 1
        if comparison:
            qt_message_dialog('PDF comparison report not supported right now')
            return

        session_infos, info = sessionutils._merge_session_info(sessions)
        if info is None:
            qt_message_dialog('Patient files do not match. Sessions may be '
                              'from different patients. Continuing without '
                              'patient info.')
            info = sessionutils.default_info()
        else:
            prompt_ = 'Please give additional subject information:'
            dlg_info = PdfReportDialog(info, prompt=prompt_)
            if dlg_info.exec_():
                new_info = dict(hetu=dlg_info.hetu, fullname=dlg_info.fullname,
                                report_notes=dlg_info.session_description)
                info.update(new_info)

                # update info files according to user input
                for session in sessions:
                    update_dict = dict(fullname=dlg_info.fullname,
                                       hetu=dlg_info.hetu)
                    session_infos[session].update(update_dict)
                    sessionutils.save_info(session, session_infos[session])

        # create the report
        if comparison:
            self._execute(nexus_make_comparison_report.do_plot,
                          thread=True,
                          finished_func=self._enable_op_buttons,
                          sessions=sessions)
        else:
            self._execute(nexus_make_pdf_report.do_plot, thread=True,
                          finished_func=self._enable_op_buttons,
                          sessionpath=sessions[0], info=info,
                          pages=dlg_info.pages)

    def _log_message(self, msg):
        c = self.txtOutput.textCursor()
        c.movePosition(QtGui.QTextCursor.End)
        self.txtOutput.setTextCursor(c)
        self.txtOutput.insertPlainText(msg)
        self.txtOutput.ensureCursorVisible()

    def _exception(self, e):
        logger.debug('caught exception while running task')
        qt_message_dialog(_exception_msg(e))

    def _disable_op_buttons(self):
        """ Disable all operation buttons """
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
        self.setEnabled(False)  # disables whole main window
        # update display immediately in case thread gets blocked
        QtWidgets.QApplication.processEvents()

    def _enable_op_buttons(self, r):
        """Enable all operation buttons and restore cursor. Takes single
        argument to fit the _finished_func call signature (see _execute) """
        QtWidgets.QApplication.restoreOverrideCursor()
        self.setEnabled(True)

    def _tardieu(self):
        """Open the Tardieu window if it is not currently open"""
        if self._tardieuwin is None or not self._tardieuwin.isVisible():
            self._tardieuwin = _tardieu.TardieuWindow()
            self._tardieuwin.show()

    def _execute(self, fun, thread=False, block_ui=True, finished_func=None,
                 **kwargs):
        """ Run function fun. If thread==True, run it in a separate worker
        thread. If block_ui, disable the ui until worker thread is finished
        (except for messages!) Returns function return value if not threaded.
        If threaded and finished_func is given, the latter is called with the
        function return value as single argument when thread completes
        execution.
        If block_ui, finished_func should re-enable UI buttons.
        kwargs are passed to function.
        """
        fun_ = partial(fun, **kwargs)
        if block_ui:
            self._disable_op_buttons()
        if thread:
            self.runner = Runner(fun_)
            if finished_func:
                self.runner.signals.finished.connect(lambda r:
                                                     finished_func(r))
            self.runner.signals.error.connect(lambda e: self._exception(e))
            self.threadpool.start(self.runner)
            retval = None
        else:  # nonthreaded execute
            try:
                retval = fun_()
            except Exception as e:
                retval = None
                self._exception(e)
            finally:
                if block_ui:
                    self._enable_op_buttons(None)
        return retval


class RunnerSignals(QObject):
    """Need a separate class since QRunnable cannot emit signals"""
    finished = pyqtSignal(object)
    error = pyqtSignal(Exception)


class Runner(QRunnable):
    """Encapsulates threaded functions for QThreadPool"""

    def __init__(self, fun):
        super(Runner, self).__init__()
        self.fun = fun
        self.signals = RunnerSignals()
        self.retval = False  # default "return value" when exception is thrown

    def run(self):
        try:
            self.retval = self.fun()
        except Exception as e:
            self.signals.error.emit(e)
        finally:
            self.signals.finished.emit(self.retval)


def main():

    app = QtWidgets.QApplication(sys.argv)

    def my_excepthook(type_, value, tback):
        """ Custom handler for unhandled exceptions:
        report to user via GUI and terminate. """
        tb_full = u''.join(traceback.format_exception(type_, value, tback))
        qt_message_dialog('Oops! An unhandled exception was generated. '
                          'The application will be closed.\n\n %s' % tb_full)
        # dump traceback to file
        # try:
        #    with io.open(Config.traceback_file, 'w', encoding='utf-8') as f:
        #        f.write(tb_full)
        # here is a danger of infinitely looping the exception hook,
        # so try to catch any exceptions...
        # except Exception:
        #    print('Cannot dump traceback!')
        sys.__excepthook__(type_, value, tback)
        app.quit()

    sys.excepthook = my_excepthook

    logger = logging.getLogger()
    handler = QtHandler()  # log to Qt logging widget
    # handler = logging.StreamHandler()   # log to sys.stdout

    handler.setFormatter(logging.
                         Formatter("%(name)s: %(levelname)s: %(message)s"))
    handler.setFormatter(logging.Formatter("%(name)s: %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)

    # quiet down some noisy loggers
    logging.getLogger('PyQt5.uic').setLevel(logging.WARNING)
    logging.getLogger('matplotlib.font_manager').setLevel(logging.WARNING)
    logging.getLogger('matplotlib.backends.backend_pdf').setLevel(logging.WARNING)
    logging.getLogger('werkzeug').setLevel(logging.WARNING)

    gaitmenu = Gaitmenu()
    gaitmenu.show()

    nexus_status = 'Vicon Nexus is %srunning' % ('' if nexus.pid() else 'not ')
    logger.debug(nexus_status)
    app.exec_()
